"""The FCOnly storage can only be used to register replicas in a FileCatalog.

It is used to preserve LFN metadata, no actual file content will be stored.

Example Configuration for Resources/StorageElement::

  ARCHIVE-SE
  {
    BackendType = None
    FCOnly
    {
      Access = remote
      Protocol = dfc
    }
  }

The 'dfc' protocol also needs to be added to the RegistrationProtocols and WriteProtocols lists in the
Operations/DataManagement section.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import gLogger, S_OK
from DIRAC.Resources.Storage.Utilities import checkArgumentFormat
from DIRAC.Resources.Storage.StorageBase import StorageBase


class FCOnlyStorage(StorageBase):
    """MetaData Storage, only for registering LFNs, nothing else."""

    _INPUT_PROTOCOLS = ["dfc"]
    _OUTPUT_PROTOCOLS = ["dfc"]

    def __init__(self, storageName, parameters):
        """c'tor

        :param self: self reference
        :param str storageName: SE name
        """
        StorageBase.__init__(self, storageName, parameters)
        self.log = gLogger.getSubLogger(__name__)

        self.pluginName = "FCOnly"
        self.protocol = []

    def getTransportURL(self, pathDict, protocols):
        res = checkArgumentFormat(pathDict)
        if not res["OK"]:
            return res
        urls = res["Value"]
        successful = {}
        failed = {}

        for url in urls:
            successful[url] = url

        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def getURLBase(self, withWSUrl=False):
        return S_OK(self.basePath)

    def removeFile(self, path):
        successful = {}
        for lfn in path:
            successful[lfn] = True
        resDict = {"Successful": successful, "Failed": {}}
        return S_OK(resDict)



"""  The Cloud Endpoint Factory has one method that instantiates a given Cloud Endpoint
"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities import ObjectLoader
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getVMTypeConfig

__RCSID__ = "$Id$"


class EndpointFactory(object):
    def __init__(self):
        """Standard constructor"""
        self.log = gLogger.getSubLogger("EndpointFactory")

    def getCE(self, site, endpoint, image=""):

        result = getVMTypeConfig(site, endpoint, image)
        if not result["OK"]:
            return result

        return self.getCEObject(parameters=result["Value"])

    def getCEObject(self, parameters=None):
        """This method returns the CloudEndpoint instance corresponding to the supplied
        CEUniqueID.  If no corresponding CE is available, this is indicated.
        """
        if not parameters:
            parameters = {}
        ceType = parameters.get("CEType", "Cloud")
        self.log.verbose("Creating Endpoint of %s type" % ceType)
        subClassName = "%sEndpoint" % (ceType)

        objectLoader = ObjectLoader.ObjectLoader()
        result = objectLoader.loadObject("Resources.Cloud.%s" % subClassName, subClassName)
        if not result["OK"]:
            gLogger.error("Failed to load object", "%s: %s" % (subClassName, result["Message"]))
            return result

        ceClass = result["Value"]
        try:
            endpoint = ceClass(parameters)
        except Exception as x:
            msg = "EndpointFactory could not instantiate %s object: %s" % (subClassName, str(x))
            self.log.exception()
            self.log.warn(msg)
            return S_ERROR(msg)

        return S_OK(endpoint)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import unittest
from mock import MagicMock
import sys

mocked_gfal2 = MagicMock()
sys.modules["gfal2"] = mocked_gfal2

from DIRAC.Resources.Storage.GFAL2_XROOTStorage import GFAL2_XROOTStorage


class XROOTStorage_TestCase(unittest.TestCase):
    def setUp(self):
        """
        Setup
        """
        # Mock external libraries / modules not interesting for the unit test

        self.parameterDict = dict(
            Protocol="protocol",
            Path="/path",
            Host="host",
            Port="",
            SvcClass="spaceToken",
            WSPath="wspath",
        )

    def test_constructURLFromLFN(self):

        resource = GFAL2_XROOTStorage("storageName", self.parameterDict)

        resource.se = MagicMock()
        voName = "voName"
        resource.se.vo = voName
        testLFN = "/%s/path/to/filename" % voName

        # # with spaceToken
        res = resource.constructURLFromLFN(testLFN)
        self.assertTrue(res["OK"])
        self.assertEqual(
            "protocol://host//path%s?svcClass=%s" % (testLFN, self.parameterDict["SvcClass"]), res["Value"]
        )

        # # no spaceToken
        resource.protocolParameters["SvcClass"] = ""
        res = resource.constructURLFromLFN(testLFN)
        self.assertTrue(res["OK"])
        self.assertEqual("protocol://host//path%s" % (testLFN,), res["Value"])


if __name__ == "__main__":

    suite = unittest.defaultTestLoader.loadTestsFromTestCase(XROOTStorage_TestCase)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/env python
"""
Script that dumps the DB information for the elements into the standard output.
If returns information concerning the StatusType and Status attributes.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import gLogger, exit as DIRACExit, version
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.ResourceStatusSystem.Client import ResourceStatusClient
from DIRAC.Core.Utilities.PrettyPrint import printTable

__RCSID__ = "$Id$"

subLogger = None
switchDict = {}


def registerSwitches():
    """
    Registers all switches that can be used while calling the script from the
    command line interface.
    """

    switches = (
        ("element=", "Element family to be Synchronized ( Site, Resource or Node )"),
        ("elementType=", "ElementType narrows the search; None if default"),
        ("name=", "ElementName; None if default"),
        ("tokenOwner=", "Owner of the token; None if default"),
        ("statusType=", "StatusType; None if default"),
        ("status=", "Status; None if default"),
        ("VO=", "Virtual organisation; None if default"),
    )

    for switch in switches:
        Script.registerSwitch("", switch[0], switch[1])


def registerUsageMessage():
    """
    Takes the script __doc__ and adds the DIRAC version to it
    """
    usageMessage = "  DIRAC %s\n" % version
    usageMessage += __doc__

    Script.setUsageMessage(usageMessage)


def parseSwitches():
    """
    Parses the arguments passed by the user
    """

    Script.parseCommandLine(ignoreErrors=True)
    args = Script.getPositionalArgs()
    if args:
        subLogger.error("Found the following positional args '%s', but we only accept switches" % args)
        subLogger.error("Please, check documentation below")
        Script.showHelp(exitCode=1)

    switches = dict(Script.getUnprocessedSwitches())
    # Default values
    switches.setdefault("elementType", None)
    switches.setdefault("name", None)
    switches.setdefault("tokenOwner", None)
    switches.setdefault("statusType", None)
    switches.setdefault("status", None)
    switches.setdefault("VO", None)

    if "element" not in switches:
        subLogger.error("element Switch missing")
        subLogger.error("Please, check documentation below")
        Script.showHelp(exitCode=1)

    if not switches["element"] in ("Site", "Resource", "Node"):
        subLogger.error("Found %s as element switch" % switches["element"])
        subLogger.error("Please, check documentation below")
        Script.showHelp(exitCode=1)

    subLogger.debug("The switches used are:")
    map(subLogger.debug, switches.items())

    return switches


def getElements():
    """
    Given the switches, gets a list of elements with their respective statustype
    and status attributes.
    """

    rssClient = ResourceStatusClient.ResourceStatusClient()

    meta = {"columns": []}
    for key in ("Name", "StatusType", "Status", "ElementType", "TokenOwner"):
        # Transforms from upper lower case to lower upper case
        if switchDict[key[0].lower() + key[1:]] is None:
            meta["columns"].append(key)

    elements = rssClient.selectStatusElement(
        switchDict["element"],
        "Status",
        name=switchDict["name"].split(",") if switchDict["name"] else None,
        statusType=switchDict["statusType"].split(",") if switchDict["statusType"] else None,
        status=switchDict["status"].split(",") if switchDict["status"] else None,
        elementType=switchDict["elementType"].split(",") if switchDict["elementType"] else None,
        tokenOwner=switchDict["tokenOwner"].split(",") if switchDict["tokenOwner"] else None,
        meta=meta,
    )

    return elements


def tabularPrint(elementsList):
    """
    Prints the list of elements on a tabular
    """

    subLogger.notice("")
    subLogger.notice("Selection parameters:")
    subLogger.notice("  %s: %s" % ("element".ljust(15), switchDict["element"]))
    titles = []
    for key in ("Name", "StatusType", "Status", "ElementType", "TokenOwner"):

        # Transforms from upper lower case to lower upper case
        keyT = key[0].lower() + key[1:]

        if switchDict[keyT] is None:
            titles.append(key)
        else:
            subLogger.notice("  %s: %s" % (key.ljust(15), switchDict[keyT]))
    subLogger.notice("")

    subLogger.notice(printTable(titles, elementsList, printOut=False, numbering=False, columnSeparator=" | "))


def run():
    """
    Main function of the script
    """

    elements = getElements()
    if not elements["OK"]:
        subLogger.error(elements)
        DIRACExit(1)
    elements = elements["Value"]

    tabularPrint(elements)


@Script()
def main():
    global subLogger
    global switchDict
    subLogger = gLogger.getSubLogger(__file__)

    # Script initialization
    registerSwitches()
    registerUsageMessage()
    switchDict = parseSwitches()

    # Run script
    run()

    # Bye
    DIRACExit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Add or Modify a Group info in DIRAC

Example:
  $ dirac-admin-add-group -G dirac_test
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

# pylint: disable=wrong-import-position

import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script

groupName = None
groupProperties = []
userNames = []


def setGroupName(arg):
    global groupName
    if groupName or not arg:
        Script.showHelp(exitCode=1)
    groupName = arg


def addUserName(arg):
    global userNames
    if not arg:
        Script.showHelp(exitCode=1)
    if arg not in userNames:
        userNames.append(arg)


def addProperty(arg):
    global groupProperties
    if not arg:
        Script.showHelp(exitCode=1)
    if arg not in groupProperties:
        groupProperties.append(arg)


@Script()
def main():
    global groupName
    global groupProperties
    global userNames
    Script.registerSwitch("G:", "GroupName:", "Name of the Group (Mandatory)", setGroupName)
    Script.registerSwitch(
        "U:", "UserName:", "Short Name of user to be added to the Group (Allow Multiple instances or None)", addUserName
    )
    Script.registerSwitch(
        "P:", "Property:", "Property to be added to the Group (Allow Multiple instances or None)", addProperty
    )
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(
        ["Property=<Value>: Other properties to be added to the Group like (VOMSRole=XXXX)"], mandatory=False
    )

    _, args = Script.parseCommandLine(ignoreErrors=True)

    if groupName is None:
        Script.showHelp(exitCode=1)

    from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

    diracAdmin = DiracAdmin()
    exitCode = 0
    errorList = []

    groupProps = {}
    if userNames:
        groupProps["Users"] = ", ".join(userNames)
    if groupProperties:
        groupProps["Properties"] = ", ".join(groupProperties)

    for prop in args:
        pl = prop.split("=")
        if len(pl) < 2:
            errorList.append(("in arguments", "Property %s has to include a '=' to separate name from value" % prop))
            exitCode = 255
        else:
            pName = pl[0]
            pValue = "=".join(pl[1:])
            gLogger.info("Setting property %s to %s" % (pName, pValue))
            groupProps[pName] = pValue

    if not diracAdmin.csModifyGroup(groupName, groupProps, createIfNonExistant=True)["OK"]:
        errorList.append(("add group", "Cannot register group %s" % groupName))
        exitCode = 255
    else:
        result = diracAdmin.csCommitChanges()
        if not result["OK"]:
            errorList.append(("commit", result["Message"]))
            exitCode = 255

    for error in errorList:
        gLogger.error("%s: %s" % error)

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



""" JobReport class encapsulates various methods of the job status reporting.
    It's an interface to JobStateUpdateClient, used when bulk submission is needed.
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from collections import defaultdict

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities import Time, DEncode
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.WorkloadManagementSystem.Client.JobStateUpdateClient import JobStateUpdateClient

__RCSID__ = "$Id$"


class JobReport(object):
    """
    .. class:: JobReport
    """

    def __init__(self, jobid, source=""):
        """c'tor"""
        self.jobStatusInfo = []  # where job status updates are cumulated
        self.appStatusInfo = []  # where application status updates are cumulated
        self.jobParameters = []  # where job parameters are cumulated
        self.jobID = int(jobid)
        self.source = source
        if not source:
            self.source = "Job_%d" % self.jobID

    def setJob(self, jobID):
        """Set the job ID for which to send reports"""
        self.jobID = jobID

    def setJobStatus(self, status="", minorStatus="", applicationStatus="", sendFlag=True):
        """Send job status information to the JobState service for jobID"""

        timeStamp = Time.toString()
        # add job status record
        self.jobStatusInfo.append((status.replace("'", ""), minorStatus.replace("'", ""), timeStamp))
        if applicationStatus:
            self.appStatusInfo.append((applicationStatus.replace("'", ""), timeStamp))
        if sendFlag and self.jobID:
            # and send
            return self.sendStoredStatusInfo()

        return S_OK()

    def setApplicationStatus(self, appStatus, sendFlag=True):
        """Send application status information to the JobState service for jobID"""
        timeStamp = Time.toString()
        # add Application status record
        if not isinstance(appStatus, str):
            appStatus = repr(appStatus)
        self.appStatusInfo.append((appStatus.replace("'", ""), timeStamp))
        if sendFlag and self.jobID:
            # and send
            return self.sendStoredStatusInfo()

        return S_OK()

    def setJobParameter(self, par_name, par_value, sendFlag=True):
        """Set job parameter for jobID"""
        self.jobParameters.append((par_name, par_value))
        if sendFlag and self.jobID:
            # and send
            return self.sendStoredJobParameters()

        return S_OK()

    def setJobParameters(self, parameters, sendFlag=True):
        """Set job parameters for jobID"""
        for pname, pvalue in parameters:
            self.jobParameters.append((pname, pvalue))

        if sendFlag and self.jobID:
            # and send
            return self.sendStoredJobParameters()

        return S_OK()

    def sendStoredStatusInfo(self):
        """Send the job status information stored in the internal cache"""

        statusDict = defaultdict(lambda: {"Source": self.source})
        for status, minor, dtime in self.jobStatusInfo:
            # No need to send empty items in dictionary
            if status:
                statusDict[dtime]["Status"] = status
            if minor:
                statusDict[dtime]["MinorStatus"] = minor
        for appStatus, dtime in self.appStatusInfo:
            # No need to send empty items in dictionary
            if appStatus:
                statusDict[dtime]["ApplicationStatus"] = appStatus

        if statusDict:
            result = JobStateUpdateClient().setJobStatusBulk(self.jobID, dict(statusDict), False)
            if result["OK"]:
                # Empty the internal status containers
                self.jobStatusInfo = []
                self.appStatusInfo = []
            return result

        else:
            return S_OK("Empty")

    def sendStoredJobParameters(self):
        """Send the job parameters stored in the internal cache"""

        if self.jobParameters:
            result = JobStateUpdateClient().setJobParameters(self.jobID, self.jobParameters)
            if result["OK"]:
                # Empty the internal parameter container
                self.jobParameters = []
            return result
        else:
            return S_OK("Empty")

    def commit(self):
        """Send all the accumulated information"""

        success = True
        result = self.sendStoredStatusInfo()
        success &= result["OK"]
        result = self.sendStoredJobParameters()
        success &= result["OK"]

        if success:
            return S_OK()
        return S_ERROR("Information upload to JobStateUpdate service failed")

    def dump(self):
        """Print out the contents of the internal cached information"""

        print("Job status info:")
        for status, minor, timeStamp in self.jobStatusInfo:
            print(status.ljust(20), minor.ljust(30), timeStamp)

        print("Application status info:")
        for status, timeStamp in self.appStatusInfo:
            print(status.ljust(20), timeStamp)

        print("Job parameters:")
        for pname, pvalue in self.jobParameters:
            print(pname.ljust(20), pvalue.ljust(30))

    def generateForwardDISET(self):
        """Generate and return failover requests for the operations in the internal cache"""
        forwardDISETOp = None

        result = self.sendStoredStatusInfo()
        if not result["OK"]:
            gLogger.error("Error while sending the job status", result["Message"])
            if "rpcStub" in result:

                rpcStub = result["rpcStub"]

                forwardDISETOp = Operation()
                forwardDISETOp.Type = "ForwardDISET"
                forwardDISETOp.Arguments = DEncode.encode(rpcStub)

            else:
                return S_ERROR("Could not create ForwardDISET operation")

        return S_OK(forwardDISETOp)



""" Base class for all services
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import six
import time
import psutil

import DIRAC

from DIRAC.Core.DISET.private.FileHelper import FileHelper
from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR, isReturnStructure
from DIRAC.Core.Utilities import Time
from DIRAC.ConfigurationSystem.Client.Config import gConfig
from DIRAC.FrameworkSystem.Client.Logger import gLogger
from DIRAC.Core.Security.Properties import CS_ADMINISTRATOR


def getServiceOption(serviceInfo, optionName, defaultValue):
    """Get service option resolving default values from the master service"""
    if optionName[0] == "/":
        return gConfig.getValue(optionName, defaultValue)
    for csPath in serviceInfo["csPaths"]:
        result = gConfig.getOption(
            "%s/%s"
            % (
                csPath,
                optionName,
            ),
            defaultValue,
        )
        if result["OK"]:
            return result["Value"]
    return defaultValue


class ConnectionError(Exception):
    def __init__(self, msg):
        self.__msg = msg

    def __str__(self):
        return "ConnectionError: %s" % self.__msg


class RequestHandler(object):
    def __init__(self, handlerInitDict, trid):
        """
        Constructor

        :type handlerInitDict: dictionary
        :param handlerInitDict: Information vars for the service

        :type trid: object
        :param trid: Transport to use
        """
        # Initially serviceInfoDict is the one base to the RequestHandler
        # the one created in _rh_initializeClass
        # FSM help me for I have made a complex stuff that I will forget in 5 mins :P
        handlerInitDict.update(self.__srvInfoDict)
        self.serviceInfoDict = handlerInitDict
        self.__trid = trid

    def initialize(self):
        """Initialize this instance of the handler (to be overwritten)"""
        pass

    @classmethod
    def _rh__initializeClass(cls, serviceInfoDict, lockManager, msgBroker, monitor):
        """
        Class initialization (not to be called by hand or overwritten!!)

        :type serviceInfoDict: dictionary
        :param serviceInfoDict: Information vars for the service
        :type msgBroker: object
        :param msgBroker: Message delivery
        :type lockManager: object
        :param lockManager: Lock manager to use
        """
        cls.__srvInfoDict = serviceInfoDict
        cls.__svcName = cls.__srvInfoDict["serviceName"]
        cls.__lockManager = lockManager
        cls.__msgBroker = msgBroker
        cls.__trPool = msgBroker.getTransportPool()
        cls.__monitor = monitor
        cls.log = gLogger

    def getRemoteAddress(self):
        """
        Get the address of the remote peer.

        :return: Address of remote peer.
        """
        return self.__trPool.get(self.__trid).getRemoteAddress()

    def getRemoteCredentials(self):
        """
        Get the credentials of the remote peer.

        :return: Credentials dictionary of remote peer.
        """
        return self.__trPool.get(self.__trid).getConnectingCredentials()

    @classmethod
    def getCSOption(cls, optionName, defaultValue=False):
        """
        Get an option from the CS section of the services

        :return: Value for serviceSection/optionName in the CS being defaultValue the default
        """
        return cls.srv_getCSOption(optionName, defaultValue)

    def _rh_executeAction(self, proposalTuple):
        """
        Execute an action.

        :type proposalTuple: tuple
        :param proposalTuple: Type of action to execute. First position of the tuple must be the type
                            of action to execute. The second position is the action itself.
        """
        actionTuple = proposalTuple[1]
        gLogger.debug("Executing %s:%s action" % tuple(actionTuple))
        startTime = time.time()
        actionType = actionTuple[0]
        self.serviceInfoDict["actionTuple"] = actionTuple
        try:
            if actionType == "RPC":
                retVal = self.__doRPC(actionTuple[1])
            elif actionType == "FileTransfer":
                retVal = self.__doFileTransfer(actionTuple[1])
            elif actionType == "Connection":
                retVal = self.__doConnection(actionTuple[1])
            else:
                return S_ERROR("Unknown action %s" % actionType)
        except ConnectionError as excp:
            gLogger.error("ConnectionError", str(excp))
            return S_ERROR(excp)
        if not isReturnStructure(retVal):
            message = "Method %s for action %s does not return a S_OK/S_ERROR!" % (actionTuple[1], actionTuple[0])
            gLogger.error(message)
            retVal = S_ERROR(message)
        elapsedTime = time.time() - startTime
        self.__logRemoteQueryResponse(retVal, elapsedTime)
        # Strip the exception/callstack info from S_ERROR responses
        if isinstance(retVal, dict):
            # ExecInfo comes from the exception
            if "ExecInfo" in retVal:
                del retVal["ExecInfo"]
            # CallStack comes from the S_ERROR construction
            if "CallStack" in retVal:
                del retVal["CallStack"]
        result = self.__trPool.send(self.__trid, retVal)  # this will delete the value from the S_OK(value)
        del retVal
        return S_OK([result, elapsedTime])

    #####
    #
    # File to/from Server Methods
    #
    #####

    def __doFileTransfer(self, sDirection):
        """
        Execute a file transfer action

        :type sDirection: string
        :param sDirection: Direction of the transfer
        :return: S_OK/S_ERROR
        """
        retVal = self.__trPool.receive(self.__trid)
        if not retVal["OK"]:
            raise ConnectionError(
                "Error while receiving file description %s %s"
                % (self.srv_getFormattedRemoteCredentials(), retVal["Message"])
            )

        # Reconvert to tuple
        fileInfo = tuple(retVal["Value"])
        sDirection = "%s%s" % (sDirection[0].lower(), sDirection[1:])
        if "transfer_%s" % sDirection not in dir(self):
            self.__trPool.send(self.__trid, S_ERROR("Service can't transfer files %s" % sDirection))
            return
        retVal = self.__trPool.send(self.__trid, S_OK("Accepted"))
        if not retVal["OK"]:
            return retVal
        self.__logRemoteQuery("FileTransfer/%s" % sDirection, fileInfo)

        self.__lockManager.lock("FileTransfer/%s" % sDirection)
        try:
            try:
                fileHelper = FileHelper(self.__trPool.get(self.__trid))
                if sDirection == "fromClient":
                    fileHelper.setDirection("fromClient")
                    uRetVal = self.transfer_fromClient(fileInfo[0], fileInfo[1], fileInfo[2], fileHelper)
                elif sDirection == "toClient":
                    fileHelper.setDirection("toClient")
                    uRetVal = self.transfer_toClient(fileInfo[0], fileInfo[1], fileHelper)
                elif sDirection == "bulkFromClient":
                    fileHelper.setDirection("fromClient")
                    uRetVal = self.transfer_bulkFromClient(fileInfo[0], fileInfo[1], fileInfo[2], fileHelper)
                elif sDirection == "bulkToClient":
                    fileHelper.setDirection("toClient")
                    uRetVal = self.transfer_bulkToClient(fileInfo[0], fileInfo[1], fileHelper)
                elif sDirection == "listBulk":
                    fileHelper.setDirection("toClient")
                    uRetVal = self.transfer_listBulk(fileInfo[0], fileInfo[1], fileHelper)
                else:
                    return S_ERROR("Direction %s does not exist!!!" % sDirection)
                if uRetVal["OK"] and not fileHelper.finishedTransmission():
                    gLogger.error("You haven't finished receiving/sending the file", str(fileInfo))
                    return S_ERROR("Incomplete transfer")
                del fileHelper
                return uRetVal
            finally:
                self.__lockManager.unlock("FileTransfer/%s" % sDirection)

        except Exception as e:  # pylint: disable=broad-except
            gLogger.exception("Uncaught exception when serving Transfer", "%s" % sDirection, lException=e)
            return S_ERROR("Server error while serving %s: %s" % (sDirection, repr(e)))

    def transfer_fromClient(self, fileId, token, fileSize, fileHelper):  # pylint: disable=unused-argument
        return S_ERROR("This server does no allow receiving files")

    def transfer_toClient(self, fileId, token, fileHelper):  # pylint: disable=unused-argument
        return S_ERROR("This server does no allow sending files")

    def transfer_bulkFromClient(self, bulkId, token, bulkSize, fileHelper):  # pylint: disable=unused-argument
        return S_ERROR("This server does no allow bulk receiving")

    def transfer_bulkToClient(self, bulkId, token, fileHelper):  # pylint: disable=unused-argument
        return S_ERROR("This server does no allow bulk sending")

    def transfer_listBulk(self, bulkId, token, fileHelper):  # pylint: disable=unused-argument
        return S_ERROR("This server does no allow bulk listing")

    #####
    #
    # RPC Methods
    #
    #####

    def __doRPC(self, method):
        """
        Execute an RPC action

        :type method: string
        :param method: Method to execute
        :return: S_OK/S_ERROR
        """
        retVal = self.__trPool.receive(self.__trid)
        if not retVal["OK"]:
            raise ConnectionError(
                "Error while receiving arguments %s %s" % (self.srv_getFormattedRemoteCredentials(), retVal["Message"])
            )
        args = retVal["Value"]
        self.__logRemoteQuery("RPC/%s" % method, args)
        return self.__RPCCallFunction(method, args)

    def __RPCCallFunction(self, method, args):
        """
        Check the arguments then call the RPC function

        :type method: string
        :param method: arguments sended by remote client

        :return: S_OK/S_ERROR
        """
        realMethod = "export_%s" % method
        gLogger.debug("RPC to %s" % realMethod)
        try:
            # Get the method we are trying to call
            oMethod = getattr(self, realMethod)
        except Exception:
            return S_ERROR("Unknown method %s" % method)
        # Check if the client sends correct arguments
        dRetVal = self.__checkExpectedArgumentTypes(method, args)
        if not dRetVal["OK"]:
            return dRetVal
        # Lock the method with Semaphore to avoid too many calls at the same time
        self.__lockManager.lock("RPC/%s" % method)
        # 18.02.19 WARNING CHRIS
        # The line bellow adds the current transportID to the message broker
        # First of all, I do not see why it is doing so.
        # Second, this affects only one every other socket, since the
        # message broker selects on that one, and in the meantime, many sockets
        # are added and removed, without even being seen by the message broker.
        # Finally, there seem to be a double read on the socket: from the message broker
        # and from the ServiceReactor, resulting in conflict.
        # This is warned in the man page of "select".
        # it has been exhibited when testing M2Crypto.
        # I will comment it out, and try to put it in a separate commit when merging
        # self.__msgBroker.addTransportId(self.__trid,
        #                                 self.serviceInfoDict['serviceName'],
        #                                 idleRead=True)
        try:
            try:
                # Trying to execute the method
                uReturnValue = oMethod(*args)
                return uReturnValue
            finally:
                # Unlock method
                self.__lockManager.unlock("RPC/%s" % method)
                # 18.02.19 WARNING CHRIS
                # See comment above
                # self.__msgBroker.removeTransport(self.__trid, closeTransport=False)
        except Exception as e:
            gLogger.exception("Uncaught exception when serving RPC", "Function %s" % method, lException=e)
            return S_ERROR("Server error while serving %s: %s" % (method, str(e)))

    def __checkExpectedArgumentTypes(self, method, args):
        """
        Check that the arguments received match the ones expected

        :type method: string
        :param method: Method to check against
        :type args: tuple
        :param args: Arguments to check
        :return: S_OK/S_ERROR
        """
        sListName = "types_%s" % method
        try:
            oTypesList = getattr(self, sListName)
        except Exception:
            gLogger.error("There's no types info for method", "export_%s" % method)
            return S_ERROR(
                "Handler error for server %s while processing method %s" % (self.serviceInfoDict["serviceName"], method)
            )
        try:
            mismatch = False
            for iIndex in range(min(len(oTypesList), len(args))):
                # If None skip the parameter
                if oTypesList[iIndex] is None:
                    continue
                # If parameter is a list or a tuple check types inside
                elif isinstance(oTypesList[iIndex], (tuple, list)):
                    if not isinstance(args[iIndex], tuple(oTypesList[iIndex])):
                        mismatch = True
                # else check the parameter
                elif not isinstance(args[iIndex], oTypesList[iIndex]):
                    mismatch = True
                # Has there been a mismatch?
                if mismatch:
                    sError = "Type mismatch in parameter %d (starting with param 0) Received %s, expected %s" % (
                        iIndex,
                        type(args[iIndex]),
                        str(oTypesList[iIndex]),
                    )
                    return S_ERROR(sError)
            if len(args) < len(oTypesList):
                return S_ERROR("Function %s expects at least %s arguments" % (method, len(oTypesList)))
        except Exception as v:
            sError = "Error in parameter check: %s" % str(v)
            gLogger.exception(sError)
            return S_ERROR(sError)
        return S_OK()

    ####
    #
    #  Connection methods
    #
    ####

    __connectionCallbackTypes = {"new": [six.string_types, dict], "connected": [], "drop": []}

    def __doConnection(self, methodName):
        """
        Connection callbacks
        """
        retVal = self.__trPool.receive(self.__trid)
        if not retVal["OK"]:
            raise ConnectionError(
                "Error while receiving arguments %s %s" % (self.srv_getFormattedRemoteCredentials(), retVal["Message"])
            )
        args = retVal["Value"]
        return self._rh_executeConnectionCallback(methodName, args)

    def _rh_executeConnectionCallback(self, methodName, args=False):
        self.__logRemoteQuery("Connection/%s" % methodName, args)
        if methodName not in RequestHandler.__connectionCallbackTypes:
            return S_ERROR("Invalid connection method %s" % methodName)
        cbTypes = RequestHandler.__connectionCallbackTypes[methodName]
        if args:
            if len(args) != len(cbTypes):
                return S_ERROR("Expected %s arguments" % len(cbTypes))
            for i in range(len(cbTypes)):
                if not isinstance(args[i], cbTypes[i]):
                    return S_ERROR("Invalid type for argument %s" % i)
            self.__trPool.associateData(self.__trid, "connectData", args)

        if not args:
            args = self.__trPool.getAssociatedData(self.__trid, "connectData")

        realMethod = "conn_%s" % methodName
        gLogger.debug("Callback to %s" % realMethod)
        try:
            oMethod = getattr(self, realMethod)
        except Exception:
            # No callback defined by handler
            return S_OK()
        try:
            if args:
                uReturnValue = oMethod(self.__trid, *args)
            else:
                uReturnValue = oMethod(self.__trid)
            return uReturnValue
        except Exception as e:
            gLogger.exception("Uncaught exception when serving Connect", "Function %s" % realMethod, lException=e)
            return S_ERROR("Server error while serving %s: %s" % (methodName, str(e)))

    def _rh_executeMessageCallback(self, msgObj):
        msgName = msgObj.getName()
        if not self.__msgBroker.getMsgFactory().messageExists(self.__svcName, msgName):
            return S_ERROR("Unknown message %s" % msgName)
        methodName = "msg_%s" % msgName
        self.__logRemoteQuery("Message/%s" % methodName, msgObj.dumpAttrs())
        startTime = time.time()
        try:
            oMethod = getattr(self, methodName)
        except Exception:
            return S_ERROR("Handler function for message %s does not exist!" % msgName)
        self.__lockManager.lock(methodName)
        try:
            try:
                uReturnValue = oMethod(msgObj)
            except Exception as e:
                gLogger.exception("Uncaught exception when serving message", methodName, lException=e)
                return S_ERROR("Server error while serving %s: %s" % (msgName, str(e)))
        finally:
            self.__lockManager.unlock(methodName)
        if not isReturnStructure(uReturnValue):
            gLogger.error("Message does not return a S_OK/S_ERROR", msgName)
            uReturnValue = S_ERROR("Message %s does not return a S_OK/S_ERROR" % msgName)
        elapsedTime = time.time() - startTime
        self.__logRemoteQueryResponse(uReturnValue, elapsedTime)
        return S_OK([uReturnValue, elapsedTime])

    ####
    #
    #  Auth methods
    #
    ####

    # @classmethod
    # def __authQuery( cls, method ):
    #  """
    #  Check if connecting user is allowed to perform an action
    #
    #  :type method: string
    #  :param method: Method to check
    #  :return: S_OK/S_ERROR
    #  """
    #  return cls.__srvInfoDict[ 'authManager' ].authQuery( method, cls.getRemoteCredentials() )

    def __logRemoteQuery(self, method, args):
        """
        Log the contents of a remote query

        :type method: string
        :param method: Method to log
        :type args: tuple
        :param args: Arguments of the method called
        """
        if self.srv_getCSOption("MaskRequestParams", True):
            argsString = "<masked>"
        else:
            argsString = "\n\t%s\n" % ",\n\t".join([str(arg)[:50] for arg in args])
        gLogger.notice("Executing action", "%s %s(%s)" % (self.srv_getFormattedRemoteCredentials(), method, argsString))

    def __logRemoteQueryResponse(self, retVal, elapsedTime):
        """
        Log the result of a query

        :type retVal: dictionary
        :param retVal: Return value of the query
        """
        if retVal["OK"]:
            argsString = "OK"
        else:
            argsString = "ERROR: %s" % retVal["Message"]
        gLogger.notice(
            "Returning response",
            "%s (%.2f secs) %s" % (self.srv_getFormattedRemoteCredentials(), elapsedTime, argsString),
        )

    ####
    #
    #  Default ping method
    #
    ####

    types_ping = []
    auth_ping = ["all"]

    def export_ping(self):
        dInfo = {}
        dInfo["version"] = DIRAC.version
        dInfo["time"] = Time.dateTime()
        # Uptime
        dInfo["host uptime"] = int(time.time() - psutil.boot_time())
        startTime = self.serviceInfoDict["serviceStartTime"]
        dInfo["service start time"] = self.serviceInfoDict["serviceStartTime"]
        serviceUptime = Time.dateTime() - startTime
        dInfo["service uptime"] = serviceUptime.days * 3600 + serviceUptime.seconds
        # Load average
        dInfo["load"] = " ".join([str(lx) for lx in os.getloadavg()])
        dInfo["name"] = self.serviceInfoDict["serviceName"]
        stTimes = os.times()
        dInfo["cpu times"] = {
            "user time": stTimes[0],
            "system time": stTimes[1],
            "children user time": stTimes[2],
            "children system time": stTimes[3],
            "elapsed real time": stTimes[4],
        }

        return S_OK(dInfo)

    types_whoami = []
    auth_whoami = ["all"]

    def export_whoami(self):
        """
        A simple whoami, returns all credential dictionary, except certificate chain object.
        """
        credDict = self.srv_getRemoteCredentials()
        if "x509Chain" in credDict:
            del credDict["x509Chain"]
        return S_OK(credDict)

    types_echo = [six.string_types]

    @staticmethod
    def export_echo(data):
        """
        This method is used for testing performance of the service

        :param str data: data to be sent back to the caller

        :return: S_OK, Value is the input data
        """
        return S_OK(data)

    types_refreshConfiguration = [bool]
    auth_refreshConfiguration = [CS_ADMINISTRATOR]

    @staticmethod
    def export_refreshConfiguration(fromMaster):
        """
        Force refreshing the configuration data

        :param bool fromMaster: flag to refresh from the master configuration service
        """
        return gConfig.forceRefresh(fromMaster=fromMaster)

    ####
    #
    #  Utilities methods
    #
    ####

    def srv_getRemoteAddress(self):
        """
        Get the address of the remote peer.

        :return: Address of remote peer.
        """
        return self.__trPool.get(self.__trid).getRemoteAddress()

    def srv_getRemoteCredentials(self):
        """
        Get the credentials of the remote peer.

        :return: Credentials dictionary of remote peer.
        """
        return self.__trPool.get(self.__trid).getConnectingCredentials()

    def srv_getFormattedRemoteCredentials(self):
        tr = self.__trPool.get(self.__trid)
        if tr:
            return tr.getFormattedCredentials()
        return "unknown"

    @classmethod
    def srv_getCSOption(cls, optionName, defaultValue=False):
        """
        Get an option from the CS section of the services

        :return: Value for serviceSection/optionName in the CS being defaultValue the default
        """
        if optionName[0] == "/":
            return gConfig.getValue(optionName, defaultValue)
        for csPath in cls.__srvInfoDict["csPaths"]:
            result = gConfig.getOption(
                "%s/%s"
                % (
                    csPath,
                    optionName,
                ),
                defaultValue,
            )
            if result["OK"]:
                return result["Value"]
        return defaultValue

    def srv_getTransportID(self):
        return self.__trid

    def srv_getClientSetup(self):
        return self.serviceInfoDict["clientSetup"]

    def srv_getClientVO(self):
        return self.serviceInfoDict["clientVO"]

    def srv_getActionTuple(self):
        if "actionTuple" not in self.serviceInfoDict:
            return ("Unknown yet",)
        return self.serviceInfoDict["actionTuple"]

    def srv_getClientVersion(self):
        return self.serviceInfoDict.get("clientVersion")

    @classmethod
    def srv_getURL(cls):
        return cls.__srvInfoDict["URL"]

    @classmethod
    def srv_getServiceName(cls):
        return cls.__srvInfoDict["serviceName"]

    @classmethod
    def srv_getMonitor(cls):
        return cls.__monitor

    def srv_msgReply(self, msgObj):
        return self.__msgBroker.sendMessage(self.__trid, msgObj)

    @classmethod
    def srv_msgSend(cls, trid, msgObj):
        return cls.__msgBroker.sendMessage(trid, msgObj)

    @classmethod
    def srv_msgCreate(cls, msgName):
        return cls.__msgBroker.getMsgFactory().createMessage(cls.__svcName, msgName)

    @classmethod
    def srv_disconnectClient(cls, trid):
        return cls.__msgBroker.removeTransport(trid)

    def srv_disconnect(self, trid=None):
        if not trid:
            trid = self.srv_getTransportID()
        return self.__msgBroker.removeTransport(trid)



"""
 This is the File StorageClass, only meant to be used localy
 """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import shutil
import errno
import stat

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Resources.Storage.Utilities import checkArgumentFormat
from DIRAC.Resources.Storage.StorageBase import StorageBase
from DIRAC.Core.Utilities.Adler import fileAdler


class FileStorage(StorageBase):
    """.. class:: FileStorage

    File storage, for local file management
    """

    def __init__(self, storageName, parameters):
        """c'tor

        :param self: self reference
        :param str storageName: SE name
        :param str protocol: protocol to use
        :param str rootdir: base path for vo files
        """

        # # init base class
        StorageBase.__init__(self, storageName, parameters)
        self.log = gLogger.getSubLogger("FileStorage")
        #     self.log.setLevel( "DEBUG" )

        self.pluginName = "File"
        self.protocol = self.protocolParameters["Protocol"]

    def getURLBase(self, withWSUrl=False):
        return S_OK(self.basePath)

    def constructURLFromLFN(self, lfn, withWSUrl=False):
        """The URL of this file protocol is very special:
        it has no protocol in front, etc. So we overwrite
        the method
        """

        result = self.getURLBase(withWSUrl=withWSUrl)
        if not result["OK"]:
            return result
        urlBase = result["Value"]
        url = os.path.join(urlBase, lfn.lstrip("/"))
        return S_OK(url)

    def exists(self, path):
        """Check if the given path exists.

        :param self: self reference
        :param path: path (or list of path) on storage
        :returns: Failed dictionary: {pfn : errorMsg}
                Successful dictionary: {pfn : bool}
        """

        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        self.log.debug("FileStorage.exists: Checking the existence of %s path(s)" % len(urls))

        successful = dict((url, os.path.exists(url)) for url in urls)

        resDict = {"Failed": {}, "Successful": successful}
        return S_OK(resDict)

    #############################################################
    #
    # These are the methods for file manipulation
    #

    def isFile(self, path):
        """Check if the given path exists and it is a file

        :param self: self reference
        :param path: path (or list of path) on storage
        :returns: Successful dict {path : boolean}
                  Failed dict {path : error message }
        """
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        self.log.debug("FileStorage.isFile: Determining whether %s paths are files." % len(urls))
        successful = {}
        failed = {}

        for url in urls:
            if os.path.exists(url):
                successful[url] = os.path.isfile(url)
            else:
                failed[url] = "No such file or directory"

        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def getFile(self, path, localPath=False):
        """make a local copy of a storage :path:

        :param self: self reference
        :param str path: path  on storage
        :param localPath: if not specified, self.cwd
        :returns: Successful dict {path : size}
                 Failed dict {path : error message }
        """

        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        self.log.debug("FileStorage.getFile: Trying to download %s files." % len(urls))

        failed = {}
        successful = {}

        if not localPath:
            localPath = os.getcwd()

        for src_url in urls:
            try:
                fileName = os.path.basename(src_url)
                dest_url = os.path.join(localPath, fileName)
                shutil.copy2(src_url, dest_url)

                fileSize = os.path.getsize(dest_url)
                successful[src_url] = fileSize
            except (OSError, IOError) as ose:
                failed[src_url] = str(ose)

        return S_OK({"Failed": failed, "Successful": successful})

    def putFile(self, path, sourceSize=0):
        """Put a copy of the local file to the current directory on the
        physical storage

        :param path: dictionnary {pfn  : localFile}
        :param sourceSize: compares the size of the local and remote.
                            You obviously run into trouble if you use a list of path...
                            If the size do not match, remove the remote file
        :returns: Successful dict {path : size}
                 Failed dict {path : error message }
        """

        if not isinstance(path, dict):
            return S_ERROR(
                "FileStorage.putFile: path argument must be a dictionary (or a list of dictionary) { url : local path}"
            )

        failed = {}
        successful = {}

        for dest_url, src_file in path.items():
            try:
                dirname = os.path.dirname(dest_url)
                if not os.path.exists(dirname):
                    os.makedirs(dirname)
                shutil.copy2(src_file, dest_url)
                fileSize = os.path.getsize(dest_url)
                if sourceSize and (sourceSize != fileSize):
                    try:
                        os.unlink(dest_url)
                    except OSError as _ose:
                        pass
                    failed[dest_url] = "Source and destination file sizes do not match (%s vs %s)." % (
                        sourceSize,
                        fileSize,
                    )
                else:
                    successful[dest_url] = fileSize
            except (OSError, IOError) as ose:
                failed[dest_url] = str(ose)

        return S_OK({"Failed": failed, "Successful": successful})

    def removeFile(self, path):
        """Remove physically the file specified by its path

        A non existing file will be considered as successfully removed.

        :param path: path (or list of path) on storage
        :returns: Successful dict {path : True}
                 Failed dict {path : error message }
        """

        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        gLogger.debug("FileStorage.removeFile: Attempting to remove %s files." % len(urls))

        failed = {}
        successful = {}

        for url in urls:
            try:
                os.unlink(url)
                successful[url] = True
            except OSError as ose:
                # Removing a non existing file is a success
                if ose.errno == errno.ENOENT:
                    successful[url] = True
                else:
                    failed[url] = str(ose)
            except Exception as e:
                failed[url] = str(e)

        return S_OK({"Failed": failed, "Successful": successful})

    @staticmethod
    def __stat(path):
        """Issue a stat call and format it the dirac way, and add the checksum

        :param self: self reference
        :param path: path on the storage
        :returns: Successful S_OK(metadataDict) or S_ERROR
        """
        try:
            statInfo = os.stat(path)
            metadataDict = {}

            metadataDict["ModTime"] = statInfo.st_mtime
            metadataDict["Size"] = statInfo.st_size
            metadataDict["Mode"] = stat.S_IMODE(statInfo.st_mode)
            metadataDict["Directory"] = bool(stat.S_ISDIR(statInfo.st_mode))
            isFile = bool(stat.S_ISREG(statInfo.st_mode))
            metadataDict["File"] = isFile

            cks = ""
            if isFile:
                cks = fileAdler(path)

            metadataDict["Checksum"] = cks if cks else ""

            # FIXME: only here for compatibility with SRM until multi protocol is properly handled
            metadataDict["Cached"] = 1
            metadataDict["Migrated"] = 0
            metadataDict["Lost"] = 0
            metadataDict["Unavailable"] = 0

            metadataDict = FileStorage._addCommonMetadata(metadataDict)

        except OSError as ose:
            return S_ERROR(str(ose))

        return S_OK(metadataDict)

    def getFileMetadata(self, path):
        """Get metadata associated to the file(s)

        :param self: self reference
        :param path: path (or list of path) on storage
        :returns: Successful dict {path : metadata}
           Failed dict {path : error message }
        """

        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        failed = {}
        successful = {}
        for url in urls:
            res = self.__stat(url)
            if not res["OK"]:
                failed[url] = res["Message"]
            elif not res["Value"]["File"]:
                failed[url] = os.strerror(errno.EISDIR)
            else:
                successful[url] = res["Value"]

        return S_OK({"Failed": failed, "Successful": successful})

    def getFileSize(self, path):
        """Get the physical size of the given file

        :param self: self reference
        :param path: path (or list of path) on storage
        :returns: Successful dict {path : size}
               Failed dict {path : error message }
        """

        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        failed = {}
        successful = {}

        for url in urls:
            try:
                # We check the filesize first because if it does not exist
                # it raises an exception, while os.path.isfile just return False
                filesize = os.path.getsize(url)
                if os.path.isfile(url):
                    successful[url] = filesize
                else:
                    failed[url] = os.strerror(errno.EISDIR)
            except OSError as ose:
                failed[url] = str(ose)

        return S_OK({"Failed": failed, "Successful": successful})

    #############################################################
    #
    # These are the methods for directory manipulation
    #

    def isDirectory(self, path):
        """Check if the given path exists and it is a directory

        :param self: self reference
        :param path: path (or list of path) on storage
        :returns: Successful dict {path : boolean}
                  Failed dict {path : error message }
        """
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        successful = {}
        failed = {}

        for url in urls:
            if os.path.exists(url):
                successful[url] = os.path.isdir(url)
            else:
                failed[url] = os.strerror(errno.ENOENT)

        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @staticmethod
    def __copyDirectory(src_dir, dest_dir):
        """Copy a whole tree, and return the number of files and the size copied

        :raise OSError:

        :param src_dir: source folder
        :param dest_dir: destination folder


        :return: dictionary with 'Files' and 'Size' as key
        """

        shutil.copytree(src_dir, dest_dir)
        nbOfFiles = 0
        totalSize = 0
        for root, _dirs, files in os.walk(src_dir):
            nbOfFiles += len(files)
            totalSize += sum(os.path.getsize(os.path.join(root, fn)) for fn in files)

        return {"Files": nbOfFiles, "Size": totalSize}

    def getDirectory(self, path, localPath=False):
        """Get locally a directory from the physical storage together with all its
        files and subdirectories.

        :param path: path (or list of path) on storage
        :param localPath: local path where to store what is downloaded
        :return: successful and failed dictionaries. The keys are the pathes,
                the values are dictionary {'Files': amount of files downloaded, 'Size': amount of data downloaded}
        """
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        self.log.debug("FileStorage.getDirectory: Attempting to get local copies of %s directories." % len(urls))

        failed = {}
        successful = {}

        if not localPath:
            localPath = os.getcwd()

        for src_dir in urls:
            try:
                dirName = os.path.basename(src_dir)
                dest_dir = "%s/%s" % (localPath, dirName)
                successful[src_dir] = self.__copyDirectory(src_dir, dest_dir)

            except OSError:
                failed[src_dir] = {"Files": 0, "Size": 0}

        return S_OK({"Failed": failed, "Successful": successful})

    def putDirectory(self, path):
        """puts a or several local directory to the physical storage together with all its files and subdirectories

        :param self: self reference
        :param dict path: dictionnary {url : local dir}
        :return: successful and failed dictionaries. The keys are the pathes,
             the values are dictionary {'Files': amount of files uploaded, 'Size': amount of data uploaded}
        """
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        successful = {}
        failed = {}

        for destDir, sourceDir in urls.items():
            try:
                successful[destDir] = self.__copyDirectory(sourceDir, destDir)
            except OSError:
                failed[destDir] = {"Files": 0, "Size": 0}

        return S_OK({"Failed": failed, "Successful": successful})

    def createDirectory(self, path):
        """Make a/several new directory on the physical storage
            This method creates all the intermediate directory

        :param self: self reference
        :param str path: path (or list of path) on storage
        :returns: Successful dict {path : True}
             Failed dict {path : error message }
        """
        urls = checkArgumentFormat(path)
        if not urls["OK"]:
            return urls
        urls = urls["Value"]

        successful = {}
        failed = {}
        self.log.debug("FileStorage.createDirectory: Attempting to create %s directories." % len(urls))

        for url in urls:
            try:
                # Create all the path
                os.makedirs(url)
                successful[url] = True
            except OSError as ose:
                failed[url] = str(ose)

        return S_OK({"Failed": failed, "Successful": successful})

    def removeDirectory(self, path, recursive=False):
        """Remove a directory on the physical storage together with all its files and
         subdirectories.

         :param path: single or list of path
         :param bool recursive: if True, we recursively delete the subdir
         :return: successful and failed dictionaries. The keys are the pathes,
               the values are dictionary {'Files': amount of files deleted, 'Size': amount of data deleted}

        .. Note:: It is known that if recursive is False, the removal of a non existing directory is successful,
              while it is failed for recursive = True. That's stupid, but well... I guess I have to keep the interface
        """
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        self.log.debug("FileStorage.removeDirectory: Attempting to remove %s directories." % len(urls))

        successful = {}
        failed = {}

        for url in urls:
            if recursive:
                nbOfFiles = 0
                totalSize = 0
                # Calculate the original size
                for root, _dirs, files in os.walk(url):
                    nbOfFiles += len(files)
                    totalSize += sum(os.path.getsize(os.path.join(root, fn)) for fn in files)
                try:
                    shutil.rmtree(url)
                    successful[url] = {"FilesRemoved": nbOfFiles, "SizeRemoved": totalSize}
                except OSError as ose:
                    # if the directory does not exist, then the numbers are already correct, no need to re do
                    # the walk
                    if ose.errno != errno.ENOENT:
                        # If we only removed partially, check how much was removed
                        leftFiles = 0
                        leftSize = 0
                        for root, _dirs, files in os.walk(url):
                            leftFiles += len(files)
                            leftSize += sum(os.path.getsize(os.path.join(root, fn)) for fn in files)
                        nbOfFiles -= leftFiles
                        totalSize -= leftSize
                    failed[url] = {"FilesRemoved": nbOfFiles, "SizeRemoved": totalSize}
            # If no recursive
            else:
                try:
                    # Delete all the files
                    for child in os.listdir(url):
                        fullpath = os.path.join(url, child)
                        if os.path.isfile(fullpath):
                            os.unlink(fullpath)
                    successful[url] = True
                except OSError as ose:
                    # If we get as exception that the directory does not exist
                    # (it can only be the directory), then success
                    if ose.errno == errno.ENOENT:
                        successful[url] = True
                    else:
                        failed[url] = str(ose)

        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def listDirectory(self, path):
        """List the supplied path

         .. warning:: It is not recursive!

        :param path: single or list of url
        :return: successful and failed dictionaries. The keys are the pathes,
              the values are dictionary 'SubDirs' and 'Files'. Each are dictionaries with
              path as key and metadata as values (for Files only, SubDirs has just True as value)
        """

        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        self.log.debug("FileStorage.listDirectory: Attempting to list %s directories." % len(urls))

        successful = {}
        failed = {}

        for url in urls:
            try:
                dirs = []
                files = []
                # os.listdir returns files and directories together
                for child in os.listdir(url):
                    fullpath = os.path.join(url, child)
                    lfnPath = os.path.join("/", os.path.relpath(fullpath, self.basePath))
                    if os.path.isfile(fullpath):
                        files.append(lfnPath)
                    else:
                        dirs.append(lfnPath)

                successful[url] = {"SubDirs": dirs, "Files": files}
            except OSError as ose:
                failed[url] = str(ose)

        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def getDirectoryMetadata(self, path):
        """Get metadata associated to the directory(ies)

        :param self: self reference
        :param path: url (or list of urls) on storage
        :returns: Successful dict {path : metadata}
                 Failed dict {path : error message }
        """

        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        failed = {}
        successful = {}
        for url in urls:
            res = self.__stat(url)
            if not res["OK"]:
                failed[url] = res["Message"]
            elif not res["Value"]["Directory"]:
                failed[url] = os.strerror(errno.ENOTDIR)
            else:
                successful[url] = res["Value"]

        return S_OK({"Failed": failed, "Successful": successful})

    def getDirectorySize(self, path):
        """Get the size of the directory on the storage

        .. warning:: the size is not recursive, and does not go into subfolders

        :param self: self reference
        :param path: path (or list of path) on storage
        :returns: list of successfull and failed dictionnary, both indexed by the path

                  * In the failed, the value is the error message
                  * In the successful the values are dictionnaries :

                      * Files : amount of files in the directory
                      * Size : summed up size of files
                      * subDirs : amount of sub directories
        """
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        self.log.debug("FileStorage.getDirectorySize: Attempting to get size of %s directories." % len(urls))

        failed = {}
        successful = {}

        for url in urls:
            try:
                totSize = 0
                nbOfFiles = 0
                nbOfSubDirs = 0

                for filename in os.listdir(url):
                    fullPath = os.path.join(url, filename)

                    if os.path.isfile(fullPath):
                        nbOfFiles += 1
                        totSize += os.path.getsize(fullPath)
                    else:
                        nbOfSubDirs += 1

                successful[url] = {"Files": nbOfFiles, "Size": totSize, "SubDirs": nbOfSubDirs}
            except OSError as ose:
                failed[url] = str(ose)

        return S_OK({"Failed": failed, "Successful": successful})



########################################################################
# File: RemoveFile.py
# Author: Krzysztof.Ciba@NOSPAMgmail.com
# Date: 2013/03/25 07:44:19
########################################################################

""" :mod: RemoveFile

    ================

    .. module: RemoveFile

    :synopsis: removeFile operation handler

    .. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

    removeFile operation handler
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id $"

# #
# @file RemoveFile.py
# @author Krzysztof.Ciba@NOSPAMgmail.com
# @date 2013/03/25 07:44:27
# @brief Definition of RemoveFile class.

# # imports
import os
import re

# # from DIRAC
from DIRAC import S_OK, S_ERROR
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
from DIRAC.DataManagementSystem.Agent.RequestOperations.DMSRequestOperationsBase import DMSRequestOperationsBase
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog

from DIRAC.MonitoringSystem.Client.MonitoringReporter import MonitoringReporter

########################################################################


class RemoveFile(DMSRequestOperationsBase):
    """
    .. class:: RemoveFile

    remove file operation handler
    """

    def __init__(self, operation=None, csPath=None):
        """c'tor

        :param self: self reference
        :param Operation operation: Operation to execute
        :param str csPath: CS path for this handler
        """
        # # call base class ctor
        DMSRequestOperationsBase.__init__(self, operation, csPath)

        # # re pattern for not existing files
        self.reNotExisting = re.compile(r"(no|not) such file.*", re.IGNORECASE)

    def __call__(self):
        """action for 'removeFile' operation"""

        # The flag  'rmsMonitoring' is set by the RequestTask and is False by default.
        # Here we use 'createRMSRecord' to create the ES record which is defined inside OperationHandlerBase.
        if self.rmsMonitoring:
            self.rmsMonitoringReporter = MonitoringReporter(monitoringType="RMSMonitoring")
        else:
            # # gMonitor stuff goes here
            gMonitor.registerActivity(
                "RemoveFileAtt", "File removals attempted", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )
            gMonitor.registerActivity(
                "RemoveFileOK", "Successful file removals", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )
            gMonitor.registerActivity(
                "RemoveFileFail", "Failed file removals", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )

        # # get waiting files
        waitingFiles = self.getWaitingFilesList()
        fc = FileCatalog(self.operation.catalogList)

        res = fc.getReplicas([wf.LFN for wf in waitingFiles])
        if not res["OK"]:
            if self.rmsMonitoring:
                for status in ["Attempted", "Failed"]:
                    self.rmsMonitoringReporter.addRecord(self.createRMSRecord(status, len(waitingFiles)))
                self.rmsMonitoringReporter.commit()
            else:
                gMonitor.addMark("RemoveFileAtt")
                gMonitor.addMark("RemoveFileFail")
            return res

        # We check the status of the SE from the LFN that are successful
        # No idea what to do with the others...
        replicas = res["Value"]["Successful"]
        targetSEs = set([se for lfn in replicas for se in replicas[lfn]])

        if targetSEs:
            # Check if SEs are allowed for remove but don't fail yet the operation if SEs are always banned
            bannedTargets = self.checkSEsRSS(targetSEs, access="RemoveAccess", failIfBanned=False)
            if not bannedTargets["OK"]:
                if self.rmsMonitoring:
                    for status in ["Attempted", "Failed"]:
                        self.rmsMonitoringReporter.addRecord(self.createRMSRecord(status, len(replicas)))
                    self.rmsMonitoringReporter.commit()
                else:
                    gMonitor.addMark("RemoveFileAtt")
                    gMonitor.addMark("RemoveFileFail")
                return bannedTargets
            bannedTargets = set(bannedTargets["Value"])
        else:
            bannedTargets = set()

        # # prepare waiting file dict
        # # We take only files that have no replica at the banned SEs... If no replica, don't
        toRemoveDict = dict(
            (
                (opFile.LFN, opFile)
                for opFile in waitingFiles
                if not bannedTargets or not bannedTargets.intersection(replicas.get(opFile.LFN, []))
            )
        )
        # If some SEs are always banned, set Failed the files that cannot be removed
        if bannedTargets and "always banned" in self.operation.Error:
            for opFile in waitingFiles:
                if opFile.LFN not in toRemoveDict:
                    # Set the files that cannot be removed Failed
                    opFile.Error = self.operation.Error
                    opFile.Status = "Failed"

            if self.rmsMonitoring:
                self.rmsMonitoringReporter.addRecord(
                    self.createRMSRecord("Failed", len(waitingFiles) - len(toRemoveDict))
                )
                self.rmsMonitoringReporter.commit()

            if not toRemoveDict:
                # If there are no files that can be removed, exit, else try once to remove them anyway
                return S_OK("%s targets are always banned for removal" % ",".join(sorted(bannedTargets)))

        if toRemoveDict:
            if self.rmsMonitoring:
                self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Attempted", len(toRemoveDict)))
            else:
                gMonitor.addMark("RemoveFileAtt", len(toRemoveDict))
            # # 1st step - bulk removal
            self.log.debug("bulk removal of %s files" % len(toRemoveDict))
            bulkRemoval = self.bulkRemoval(toRemoveDict)
            if not bulkRemoval["OK"]:
                self.log.error("Bulk file removal failed", bulkRemoval["Message"])
            else:
                if self.rmsMonitoring:
                    self.rmsMonitoringReporter.addRecord(
                        self.createRMSRecord("Successful", len(toRemoveDict) - len(bulkRemoval["Value"]))
                    )
                else:
                    gMonitor.addMark("RemoveFileOK", len(toRemoveDict) - len(bulkRemoval["Value"]))

            # # 2nd step - single file removal
            for lfn, opFile in toRemoveDict.items():
                self.log.info("removing single file %s" % lfn)
                singleRemoval = self.singleRemoval(opFile)
                if not singleRemoval["OK"]:
                    self.log.error("Error removing single file", singleRemoval["Message"])
                    if self.rmsMonitoring:
                        self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Failed", 1))
                    else:
                        gMonitor.addMark("RemoveFileFail", 1)
                else:
                    self.log.info("file %s has been removed" % lfn)
                    if self.rmsMonitoring:
                        self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Successful", 1))
                    else:
                        gMonitor.addMark("RemoveFileOK", 1)

            # # set
            failedFiles = [
                (lfn, opFile) for (lfn, opFile) in toRemoveDict.items() if opFile.Status in ("Failed", "Waiting")
            ]
            if failedFiles:
                self.operation.Error = "failed to remove %d files" % len(failedFiles)

        if self.rmsMonitoring:
            self.rmsMonitoringReporter.commit()

        if bannedTargets:
            return S_OK("%s targets are banned for removal" % ",".join(sorted(bannedTargets)))
        return S_OK()

    def bulkRemoval(self, toRemoveDict):
        """bulk removal using request owner DN

        :param dict toRemoveDict: { lfn: opFile, ... }
        :return: S_ERROR or S_OK( { lfn: opFile, ... } ) -- dict with files still waiting to be removed
        """
        bulkRemoval = self.dm.removeFile(list(toRemoveDict), force=True)
        if not bulkRemoval["OK"]:
            error = bulkRemoval["Message"]
            self.log.error("Bulk file removal failed", error)
            self.operation.Error = error
            for opFile in self.operation:
                opFile.Error = error
            return bulkRemoval
        bulkRemoval = bulkRemoval["Value"]
        # # filter results
        for lfn, opFile in toRemoveDict.items():
            if lfn in bulkRemoval["Successful"]:
                opFile.Status = "Done"
            elif lfn in bulkRemoval["Failed"]:

                error = bulkRemoval["Failed"][lfn]
                if isinstance(error, dict):
                    error = ";".join(["%s-%s" % (k, v) for k, v in error.items()])
                opFile.Error = error
                if self.reNotExisting.search(opFile.Error):
                    opFile.Status = "Done"

        # # return files still waiting
        toRemoveDict = dict((lfn, opFile) for lfn, opFile in toRemoveDict.items() if opFile.Status == "Waiting")
        return S_OK(toRemoveDict)

    def singleRemoval(self, opFile):
        """remove single file

        :param opFile: File instance
        """
        # # try to remove with owner proxy
        proxyFile = None
        if "Write access not permitted for this credential" in opFile.Error:
            if "DataManager" in self.shifter:
                # #  you're a data manager - get proxy for LFN and retry
                saveProxy = os.environ["X509_USER_PROXY"]
                try:
                    fileProxy = self.getProxyForLFN(opFile.LFN)
                    if not fileProxy["OK"]:
                        opFile.Error = "Error getting owner's proxy : %s" % fileProxy["Message"]
                    else:
                        proxyFile = fileProxy["Value"]
                        self.log.info("Trying to remove file with owner's proxy (file %s)" % proxyFile)

                        removeFile = self.dm.removeFile(opFile.LFN, force=True)
                        self.log.always(str(removeFile))

                        if not removeFile["OK"]:
                            opFile.Error = str(removeFile["Message"])
                            if self.reNotExisting.search(str(removeFile["Message"]).lower()):
                                opFile.Status = "Done"
                        else:
                            removeFile = removeFile["Value"]
                            if opFile.LFN in removeFile["Failed"]:
                                error = removeFile["Failed"][opFile.LFN]
                                if isinstance(error, dict):
                                    error = ";".join(["%s-%s" % (k, v) for k, v in error.items()])
                                if self.reNotExisting.search(error):
                                    # This should never happen due to the "force" flag
                                    opFile.Status = "Done"
                                else:
                                    opFile.Error = error
                            else:
                                opFile.Status = "Done"
                finally:
                    if proxyFile:
                        os.unlink(proxyFile)
                    # # put back request owner proxy to env
                    os.environ["X509_USER_PROXY"] = saveProxy
        # # file removed? update its status to 'Done'
        if opFile.Status == "Done":
            return S_OK()
        return S_ERROR(opFile.Error)



#!/usr/bin/python

########################################################################
# File :    DIRACbenchmark.py
# Author :  Andrew McNab
########################################################################

""" DIRAC Benchmark 2012 by Ricardo Graciani, and wrapper functions to
    run multiple copies in parallel by Andrew McNab.

    This file (DIRACbenchmark.py) is intended to be the ultimate upstream
    shared by different users of the DIRAC Benchmark 2012 (DB12). The
    canonical version can be found at https://github.com/DIRACGrid/DB12

    This script can either be imported or run from the command line:

    ./DIRACbenchmark.py NUMBER

    where NUMBER gives the number of benchmark processes to run in parallel.

    Run  ./DIRACbenchmark.py help  to see more options.
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from past.builtins import long
import os
import sys
import random
import six
from six.moves.urllib.request import urlopen
import multiprocessing

version = "00.04 DB12"


def singleDiracBenchmark(iterations=1, measuredCopies=None):
    """Get Normalized Power of one CPU in DIRAC Benchmark 2012 units (DB12)"""

    # This number of iterations corresponds to 1kHS2k.seconds, i.e. 250 HS06 seconds

    n = int(1000 * 1000 * 12.5)
    calib = 250.0

    m = long(0)
    m2 = long(0)
    p = 0
    p2 = 0
    # Do one iteration extra to allow CPUs with variable speed (we ignore zeroth iteration)
    # Do one or more extra iterations to avoid tail effects when copies run in parallel
    i = 0
    while (i <= iterations) or (measuredCopies is not None and measuredCopies.value > 0):
        if i == 1:
            start = os.times()

        # Now the iterations
        for _j in six.moves.range(n):
            t = random.normalvariate(10, 1)
            m += t
            m2 += t * t
            p += t
            p2 += t * t

        if i == iterations:
            end = os.times()
            if measuredCopies is not None:
                # Reduce the total of running copies by one
                measuredCopies.value -= 1

        i += 1

    cput = sum(end[:4]) - sum(start[:4])
    wall = end[4] - start[4]

    if not cput:
        return None

    # Return DIRAC-compatible values
    return {"CPU": cput, "WALL": wall, "NORM": calib * iterations / cput, "UNIT": "DB12"}


def singleDiracBenchmarkProcess(resultObject, iterations=1, measuredCopies=None):
    """Run singleDiracBenchmark() in a multiprocessing friendly way"""

    benchmarkResult = singleDiracBenchmark(iterations=iterations, measuredCopies=measuredCopies)

    if not benchmarkResult or "NORM" not in benchmarkResult:
        return None

    # This makes it easy to use with multiprocessing.Process
    resultObject.value = benchmarkResult["NORM"]


def multipleDiracBenchmark(copies=1, iterations=1, extraIteration=False):
    """Run multiple copies of the DIRAC Benchmark in parallel"""

    processes = []
    results = []

    if extraIteration:
        # If true, then we run one or more extra iterations in each
        # copy until the number still being meausured is zero.
        measuredCopies = multiprocessing.Value("i", copies)
    else:
        measuredCopies = None

    # Set up all the subprocesses
    for i in range(copies):
        results.append(multiprocessing.Value("d", 0.0))
        processes.append(
            multiprocessing.Process(target=singleDiracBenchmarkProcess, args=(results[i], iterations, measuredCopies))
        )

    # Start them all off at the same time
    for p in processes:
        p.start()

    # Wait for them all to finish
    for p in processes:
        p.join()

    raw = []
    product = 1.0

    for result in results:
        raw.append(result.value)
        product *= result.value

    raw.sort()

    # Return the list of raw results and various averages
    return {
        "raw": raw,
        "copies": copies,
        "sum": sum(raw),
        "arithmetic_mean": sum(raw) / copies,
        "geometric_mean": product ** (1.0 / copies),
        "median": raw[(copies - 1) / 2],
    }


def wholenodeDiracBenchmark(copies=None, iterations=1, extraIteration=False):
    """Run as many copies as needed to occupy the whole machine"""

    # Try $MACHINEFEATURES first if not given by caller
    if copies is None and "MACHINEFEATURES" in os.environ:
        try:
            copies = int(urlopen(os.environ["MACHINEFEATURES"] + "/total_cpu").read())
        except Exception:
            pass

    # If not given by caller or $MACHINEFEATURES/total_cpu then just count CPUs
    if copies is None:
        try:
            copies = multiprocessing.cpu_count()
        except Exception:
            copies = 1

    return multipleDiracBenchmark(copies=copies, iterations=iterations, extraIteration=extraIteration)


def jobslotDiracBenchmark(copies=None, iterations=1, extraIteration=False):
    """Run as many copies as needed to occupy the job slot"""

    # Try $JOBFEATURES first if not given by caller
    if copies is None and "JOBFEATURES" in os.environ:
        try:
            copies = int(urlopen(os.environ["JOBFEATURES"] + "/allocated_cpu").read())
        except Exception:
            pass

    # If not given by caller or $JOBFEATURES/allocated_cpu then just run one copy
    if copies is None:
        copies = 1

    return multipleDiracBenchmark(copies=copies, iterations=iterations, extraIteration=extraIteration)


#
# If we run as a command
#
if __name__ == "__main__":

    helpString = """DIRACbenchmark.py [--iterations ITERATIONS] [--extra-iteration]
                  [COPIES|single|wholenode|jobslot|version|help]

Uses the functions within DIRACbenchmark.py to run the DB12 benchmark from the
command line.

By default one benchmarking iteration is run, in addition to the initial
iteration which DB12 runs and ignores to avoid ramp-up effects at the start.
The number of benchmarking iterations can be increased using the --iterations
option. Additional iterations which are also ignored can be added with the
--extra-iteration option  to avoid tail effects. In this case copies which
finish early run additional iterations until all the measurements finish.

The COPIES (ie an integer) argument causes multiple copies of the benchmark to
be run in parallel. The tokens "wholenode", "jobslot" and "single" can be
given instead to use $MACHINEFEATURES/total_cpu, $JOBFEATURES/allocated_cpu,
or 1 as the number of copies respectively. If $MACHINEFEATURES/total_cpu is
not available, then the number of (logical) processors visible to the
operating system is used.

Unless the token "single" is used, the script prints the following results to
two lines on stdout:

COPIES SUM ARITHMETIC-MEAN GEOMETRIC-MEAN MEDIAN
RAW-RESULTS

The tokens "version" and "help" print information about the script.

The source code of DIRACbenchmark.py provides examples of how the functions
within DIRACbenchmark.py can be used by other Python programs.

DIRACbenchmark.py is distributed from  https://github.com/DIRACGrid/DB12
"""

    copies = None
    iterations = 1
    extraIteration = False

    for arg in sys.argv[1:]:
        if arg.startswith("--iterations="):
            iterations = int(arg[13:])
        elif arg == "--extra-iteration":
            extraIteration = True
        elif arg == "--help" or arg == "help":
            print(helpString)
            sys.exit(0)
        elif not arg.startswith("--"):
            copies = arg

    if copies == "version":
        print(version)
        sys.exit(0)

    if copies is None or copies == "single":
        print(singleDiracBenchmark()["NORM"])
        sys.exit(0)

    if copies == "wholenode":
        result = wholenodeDiracBenchmark(iterations=iterations, extraIteration=extraIteration)
        print(result["copies"], result["sum"], result["arithmetic_mean"], result["geometric_mean"], result["median"])
        print(" ".join([str(i) for i in result["raw"]]))
        sys.exit(0)

    if copies == "jobslot":
        result = jobslotDiracBenchmark(iterations=iterations, extraIteration=extraIteration)
        print(result["copies"], result["sum"], result["arithmetic_mean"], result["geometric_mean"], result["median"])
        print(" ".join([str(i) for i in result["raw"]]))
        sys.exit(0)

    result = multipleDiracBenchmark(copies=int(copies), iterations=iterations, extraIteration=extraIteration)
    print(result["copies"], result["sum"], result["arithmetic_mean"], result["geometric_mean"], result["median"])
    print(" ".join([str(i) for i in result["raw"]]))
    sys.exit(0)



""" This contains unit tests to make sure that the migration between PyGSI and M2Crypto is as smooth as possible
    The test covers only the method exposed by the PyGSI version for the time being.

    We are not testing:
    -> setChain: simple setter
    -> setPKey: simple setter

    * generateProxyRequest
    * generateChainFromRequestString


    We are skipping:
    * init with certList as argument, because never used
    * generateProxyToString: strength (nah....)


    We are missing:
      test proxy from proxy ? Is it a usecase ?


"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


# redefined-outer-name is needed because we keep bassing get_X509Chain_class as param
# pylint: disable=redefined-outer-name

from datetime import datetime, timedelta
from string import ascii_letters, digits
import sys

from hypothesis import given, settings, HealthCheck
from hypothesis.strategies import integers, text

from pytest import mark, fixture, skip, raises, approx

parametrize = mark.parametrize

# function_scoped_fixture is only used in Python 3 compatible release of hypothesis
if sys.version_info.major >= 3:
    function_scoped = (HealthCheck.function_scoped_fixture,)
else:
    function_scoped = tuple()

from DIRAC.Core.Security.test.x509TestUtilities import (
    CERTS,
    CERTKEYS,
    CERTCONTENTS,
    deimportDIRAC,
    ENCRYPTEDKEYPASS,
    ENCRYPTEDKEY,
    getCertOption,
    HOSTCERT,
    KEYCONTENTS_PKCS8,
    USERCERT,
    get_X509Chain_class,
    X509CHAINTYPES,
    get_X509Request,
    get_X509Chain_from_X509Request,
)


ONE_YEAR_IN_SECS = 3600 * 24 * 365
TWENTY_YEARS_IN_SEC = 20 * ONE_YEAR_IN_SECS

# Validity date wont go further than 2050, See RFC 5280 4.1.2.5 for more information.
NO_LATER_THAN_2050_IN_SEC = int((datetime.strptime("2049-12-31", "%Y-%M-%d") - datetime.now()).total_seconds())


@fixture(scope="function", params=X509CHAINTYPES)
def get_proxy(request):
    """Fixture to return either the proxy string.
    It also 'de-import' DIRAC before and after
    """
    # Clean before

    # Oh what a dirty hack....
    # When you do the delegation, you call both Request and Proxy generation fixtures.
    # So if you do the cleaning twice, you end up in a terrible mess.
    # So, do not do the cleaning if you are in the test_delegation method
    if request.function.__name__ != "test_delegation":
        deimportDIRAC()
    x509Class = request.param

    if x509Class == "M2_X509Chain":
        from DIRAC.Core.Security.m2crypto.X509Chain import X509Chain
    else:
        raise NotImplementedError()

    def _generateProxy(certFile, lifetime=3600, **kwargs):
        """Generate the proxyString and return it as an X509Chain object

        :param certFile: path to the certificate
        :param lifetime: lifetime of the proxy in seconds

        :returns:  X509Chain object
        """
        # Load the certificate and the key
        x509Chain = X509Chain()
        x509Chain.loadChainFromFile(certFile)
        x509Chain.loadKeyFromFile(getCertOption(certFile, "keyFile"))

        # Generate the proxy string
        res = x509Chain.generateProxyToString(lifetime, rfc=True, **kwargs)

        proxyString = res["Value"]
        # Load the proxy string as an X509Chain object
        proxyChain = X509Chain()
        proxyChain.loadProxyFromString(proxyString)

        return proxyChain

    yield _generateProxy

    # Clean after
    deimportDIRAC()


@parametrize("cert_file", CERTS)
def test_loadChainFromFile(cert_file, get_X509Chain_class):
    """ " Just load a certificate chain"""
    x509Chain = get_X509Chain_class()
    res = x509Chain.loadChainFromFile(cert_file)
    assert res["OK"]


def test_loadChainFromFile_non_existing_file(get_X509Chain_class):
    """ " Just loadChain a non existing file"""
    X509Chain = get_X509Chain_class()
    res = X509Chain.loadChainFromFile("/tmp/nonexistingFile.pem")
    assert not res["OK"]

    from DIRAC.Core.Utilities.DErrno import EOF

    assert res["Errno"] == EOF


# pylint: disable=unused-argument
@parametrize("cert_content_type", CERTCONTENTS)
def test_loadChainFromString(cert_content_type, get_X509Chain_class, indirect=("hostcertcontent", "usercertcontent")):
    """ " Just loadChain a certificate from PEM string
    :param cert_content_type: either HOSTCERTCONTENT or USERCERTCONTENT

    """
    X509Chain = get_X509Chain_class()
    res = X509Chain.loadChainFromString(CERTCONTENTS[cert_content_type])
    assert res["OK"], res


def test_loadChainFromString_non_pem(get_X509Chain_class):
    """ " Just loadChain a non pem formated string"""
    X509Chain = get_X509Chain_class()
    res = X509Chain.loadChainFromString("THIS IS NOT PEM DATA")
    assert not res["OK"]

    from DIRAC.Core.Utilities.DErrno import EX509

    assert res["Errno"] == EX509


@parametrize("key_file", CERTKEYS)
def test_init_with_key(key_file, get_X509Chain_class):
    """Test init with key object as argument and check the content"""

    chain1 = get_X509Chain_class()
    chain1.loadKeyFromFile(key_file)

    # Get the key and check the number of bits
    keyObj = chain1.getPKeyObj()["Value"]

    chain2 = get_X509Chain_class(keyObj=keyObj)
    assert chain1.dumpPKeyToString() == chain2.dumpPKeyToString()
    # Careful ! The two keys are the same object
    assert chain2.getPKeyObj()["Value"] is keyObj


@parametrize("key_file", CERTKEYS)
def test_privatekey_without_password(key_file, get_X509Chain_class):
    """Test loading a key from a file, retrieve the object and check the content"""

    X509Chain = get_X509Chain_class()
    res = X509Chain.loadKeyFromFile(key_file)
    assert res["OK"]
    # Get the key and check the number of bits
    res = X509Chain.getPKeyObj()
    assert res["OK"]
    assert res["Value"].size() == 512

    # Check that the content of the object is correct.
    # CAUTION ! The object is PKCS8, while the file contains PKCS1.
    # Check the comment of KEYCONTENTS_PKCS8
    res = X509Chain.dumpPKeyToString()
    assert res["Value"] == KEYCONTENTS_PKCS8[key_file]


def test_privatekey_with_password(get_X509Chain_class):
    """Test loading a password protected key from a file and retrieve the object"""
    X509Chain = get_X509Chain_class()
    res = X509Chain.loadKeyFromFile(ENCRYPTEDKEY, password=ENCRYPTEDKEYPASS)
    assert res["OK"]
    # Get the key and check the number of bits
    res = X509Chain.getPKeyObj()
    assert res["OK"]
    assert res["Value"].size() == 512


def test_privatekey_with_wrong_password(get_X509Chain_class):
    """Try loading a password protected key with the wrong password"""
    X509Chain = get_X509Chain_class()
    res = X509Chain.loadKeyFromFile(ENCRYPTEDKEY, password="WRONGPASSWRD")
    assert not res["OK"]

    from DIRAC.Core.Utilities.DErrno import ECERTREAD

    assert res["Errno"] == ECERTREAD


@parametrize("cert_file", CERTS)
def test_getCertInChain_on_cert(cert_file, get_X509Chain_class):
    """ " Load a chain, get the first certificate, and check its name"""
    x509Chain = get_X509Chain_class()
    x509Chain.loadChainFromFile(cert_file)
    res = x509Chain.getCertInChain(0)
    assert res["OK"]

    certSubject = res["Value"].getSubjectDN().get("Value")
    assert certSubject == getCertOption(cert_file, "subjectDN")


def test_getCertInChain_too_far(get_X509Chain_class):
    """ " Load a chain, get too far in the certificate chain"""
    x509Chain = get_X509Chain_class()
    x509Chain.loadChainFromFile(HOSTCERT)
    # it should raise IndexError if too far
    with raises(IndexError):
        x509Chain.getCertInChain(1)


@parametrize("cert_file", CERTS)
def test_getCertList(cert_file, get_X509Chain_class):
    """ " Load a chain, and get its length."""
    x509Chain = get_X509Chain_class()
    x509Chain.loadChainFromFile(cert_file)
    # For a certificate, there should be only 1 certificate in the chain

    assert len(x509Chain.getCertList()["Value"]) == 1
    assert len(x509Chain.getCertList()["Value"]) == x509Chain.getNumCertsInChain()["Value"]


@parametrize("cert_file", CERTS)
def test_certProperties(cert_file, get_X509Chain_class):
    """Try on a certificate if it is a proxy, limited proxy, VOMS, valid proxy, rfc"""
    x509Chain = get_X509Chain_class()
    x509Chain.loadChainFromFile(cert_file)

    # These methods should return False
    assert x509Chain.isProxy()["Value"] is False
    assert x509Chain.isLimitedProxy()["Value"] is False
    assert x509Chain.isVOMS()["Value"] is False

    assert x509Chain.isRFC()["Value"] is False

    from DIRAC.Core.Utilities.DErrno import ENOCHAIN

    # Now these methods should complain that it is not a proxy
    # After all, why would you do something logical...
    assert x509Chain.isValidProxy()["Errno"] == ENOCHAIN


@parametrize("cert_file", CERTS)
def test_getVOMSData_on_cert(cert_file, get_X509Chain_class):
    """ " Load a  Chain with only a certificate and load the (non existing VOMS data)
    Of course, it will behave differently from the certificate...
    """

    x509Chain = get_X509Chain_class()
    x509Chain.loadChainFromFile(cert_file)

    res = x509Chain.getVOMSData()

    assert res["OK"]

    # The VOMS data of a certificate chain composed of only a certificate is... False
    assert res["Value"] is False


@parametrize("cert_file", CERTS)
def test_getDIRACGroup_on_cert(cert_file, get_X509Chain_class):
    """ " Load a  Chain with only a certificate and get the (non existing) DIRAC Group
    Of course, it will behave differently from the certificate...
    """

    x509Chain = get_X509Chain_class()
    x509Chain.loadChainFromFile(cert_file)

    # ignoreDefault is used such that there is no attempt to look for group in the CS
    res = x509Chain.getDIRACGroup(ignoreDefault=True)

    assert not res["OK"]

    from DIRAC.Core.Utilities.DErrno import EX509

    assert res["Errno"] == EX509


# TODO: have a non valid certificate to try


@parametrize("cert_file", CERTS)
def test_hasExpired(cert_file, get_X509Chain_class):
    """ " Load a valid certificate and check it has not expired"""
    x509Chain = get_X509Chain_class()
    x509Chain.loadChainFromFile(cert_file)

    res = x509Chain.hasExpired()
    assert res["OK"]
    assert not res["Value"]


@parametrize("cert_file", CERTS)
def test_getNotAfterDate(cert_file, get_X509Chain_class):
    """ " Load a valid certificate and check its expiration date"""
    x509Chain = get_X509Chain_class()
    x509Chain.loadChainFromFile(cert_file)

    res = x509Chain.getNotAfterDate()

    assert res["OK"]
    # We expect getNotAfterDate to return a datetime
    assert res["Value"].date() == getCertOption(cert_file, "endDate")


@parametrize("cert_file", CERTS)
def test_getRemainingSecs_on_cert(cert_file, get_X509Chain_class):
    """ " Load a valid certificate and check the output is a positive integer"""

    x509Chain = get_X509Chain_class()
    x509Chain.loadChainFromFile(cert_file)

    res = x509Chain.getRemainingSecs()

    assert res["OK"]
    assert isinstance(res["Value"], int) and res["Value"] > 0


@parametrize("cert_file", CERTS)
def test_dumpChainToString_on_cert(cert_file, get_X509Chain_class):
    """ " Load a valid certificate in a chain, and dump all to string"""

    x509Chain = get_X509Chain_class()
    x509Chain.loadChainFromFile(cert_file)

    res = x509Chain.dumpChainToString()

    assert res["OK"]

    assert res["Value"] == getCertOption(cert_file, "content")


@parametrize("cert_file", CERTS)
def test_isPUSP_on_cert(cert_file, get_X509Chain_class):
    """ " Load a valid certificate in a chain, and check isPUSP"""

    x509Chain = get_X509Chain_class()
    x509Chain.loadChainFromFile(cert_file)

    res = x509Chain.isPUSP()

    assert res["OK"]
    assert res["Value"] is False


@parametrize("cert_file", CERTS)
def test_getCredentials_on_cert(cert_file, get_X509Chain_class):
    """ " Load a valid certificate in a chain, and check the information returned.
    We do not check the values, they are already checked in other tests
    """

    x509Chain = get_X509Chain_class()
    x509Chain.loadChainFromFile(cert_file)

    credentialInfo = ["DN", "isLimitedProxy", "isProxy", "issuer", "secondsLeft", "subject", "validDN", "validGroup"]

    res = x509Chain.getCredentials(ignoreDefault=True)

    assert res["OK"]
    assert sorted(res["Value"]) == sorted(credentialInfo)


@parametrize("cert_file", CERTS)
def test_hash_on_cert(cert_file, get_X509Chain_class):
    """ " Load a valid certificate in a chain, and check the hash
    It is supposed to raise an exception because it is not a proxy
    """

    x509Chain = get_X509Chain_class()
    x509Chain.loadChainFromFile(cert_file)

    # Because hash expects a proxy, it will attempt to access the diracGroup attribute
    # and raise a KeyError
    with raises(KeyError):
        x509Chain.hash()


#######################################################################
# From here we start proxy tests
########################################################################


def test_generateProxyToString(get_proxy):
    """ " Generate a proxy and check that the chain has two level"""

    proxyChain = get_proxy(USERCERT)
    # We are supposed now to have two elements in the chain
    assert proxyChain.getNumCertsInChain()["Value"] == 2


def test_getCertInChain(get_proxy):
    """ " retrieve the first certificate in the chain, and make sure it is the original one"""

    proxyChain = get_proxy(USERCERT)

    chainLength = proxyChain.getNumCertsInChain()["Value"]

    res = proxyChain.getCertInChain(certPos=chainLength - 1)

    assert res["OK"]

    certSubject = res["Value"].getSubjectDN().get("Value")
    assert certSubject == getCertOption(USERCERT, "subjectDN")

    # bonus: check the negative counter also works
    assert certSubject == proxyChain.getCertInChain(certPos=-1)["Value"].getSubjectDN().get("Value")

    # Test default value
    assert proxyChain.isProxy()["Value"] is True
    assert proxyChain.isLimitedProxy()["Value"] is False
    assert proxyChain.isVOMS()["Value"] is False
    assert proxyChain.isRFC()["Value"] is True
    assert proxyChain.isValidProxy()["Value"] is True


@mark.slow
@settings(max_examples=200, suppress_health_check=function_scoped)
@given(lifetime=integers(max_value=ONE_YEAR_IN_SECS, min_value=1))
def test_proxyLifetime(get_proxy, lifetime):
    """ " Generate a proxy with various lifetime, smaller than the certificate length
    :param lifetime: lifetime of the proxy in seconds
    """

    proxyChain = get_proxy(USERCERT, lifetime=lifetime)

    res = proxyChain.getNotAfterDate()
    assert res["OK"]

    notAfterDate = res["Value"]
    expectedValidity = datetime.utcnow() + timedelta(seconds=lifetime)

    # The two value should coincide with a margin of 2 seconds
    margin = 2
    assert (notAfterDate - expectedValidity).total_seconds() == approx(0, abs=margin)


@mark.slow
@settings(max_examples=200, suppress_health_check=function_scoped)
@given(lifetime=integers(min_value=TWENTY_YEARS_IN_SEC, max_value=NO_LATER_THAN_2050_IN_SEC))
def test_tooLong_proxyLifetime(get_proxy, lifetime):
    """ " Generate a proxy with various lifetime, longer than the certificate length
    :param lifetime: lifetime of the proxy in seconds
    """

    proxyChain = get_proxy(USERCERT, lifetime=lifetime)

    res = proxyChain.getNotAfterDate()
    assert res["OK"]

    notAfterDate = res["Value"]

    # The expected validity is the validity of the certificate

    certObj = proxyChain.getCertInChain(-1)["Value"]
    expectedEndDate = certObj.getNotAfterDate()["Value"]
    # The two value should coincide with a margin of 2 seconds
    assert notAfterDate == expectedEndDate


# def generateProxyToString(self, lifeTime, diracGroup=False, strength=1024, limited=False, proxyKey=False, rfc = True):

# hypthesis successfully prove that m2crypto implementation does not work
# for an empty group name, or 0, or whatever can be evaluated to False . Fine...
# Let's just focus on letters and '-'


@mark.slow
@settings(max_examples=200, suppress_health_check=function_scoped)
@given(diracGroup=text(ascii_letters + "-_" + digits, min_size=1))
def test_diracGroup(get_proxy, diracGroup):
    """Generate a proxy with a given group and check that we can retrieve it"""
    proxyChain = get_proxy(USERCERT, diracGroup=diracGroup)

    res = proxyChain.getDIRACGroup(ignoreDefault=True)
    assert res["OK"]

    assert len(res["Value"]) == len(diracGroup)
    assert res["Value"] == diracGroup


@parametrize("isLimited", (True, False))
def test_limitedProxy(get_proxy, isLimited):
    """Generate limited and non limited proxy"""
    # A group is needed to be limited
    proxyChain = get_proxy(USERCERT, diracGroup="anyGroup", limited=isLimited)

    res = proxyChain.isLimitedProxy()
    assert res["OK"]

    assert res["Value"] is isLimited


def test_getIssuerCert(get_proxy):
    """Generate a proxy and check the issuer of the certificate"""
    proxyChain = get_proxy(USERCERT)

    res = proxyChain.getIssuerCert()
    assert res["OK"]

    assert res["Value"].getSubjectDN()["Value"] == getCertOption(USERCERT, "subjectDN")


################################################################
# From now on, test proxy coming from Requests
################################################################

#
# retVal = chain.generateChainFromRequestString(reqDict['request'],
#                                               lifetime=chainLifeTime,
#                                               diracGroup=diracGroup,
#                                               rfc = rfcIfPossible)
@mark.slow
@settings(max_examples=200, suppress_health_check=function_scoped)
@given(diracGroup=text(ascii_letters + "-", min_size=1), lifetime=integers(min_value=1, max_value=TWENTY_YEARS_IN_SEC))
def test_delegation(get_X509Request, get_proxy, diracGroup, lifetime):
    """
    Test the delegation mechanism.
    Generate a proxy request and generate the proxy from there
    NOTE: DO NOT CHANGE THE NAME OF THIS TEST FUNCTION ! See get_proxy code for details

    :param diracGroup: group of the initial proxy
    :param lifetime: requested lifetime of the delegated proxy
    """

    # The server side generates a request
    # Equivalent to ProxyManager.requestDelegationUpload
    x509Req = get_X509Request()
    x509Req.generateProxyRequest()
    reqStr = x509Req.dumpRequest()["Value"]

    # This object contains both the public and private key
    pkeyReq = x509Req.getPKey()

    #######################################################

    # The client side signs the request

    proxyChain = get_proxy(USERCERT, diracGroup=diracGroup)

    # The proxy will contain a "bullshit private key"
    res = proxyChain.generateChainFromRequestString(reqStr, lifetime=lifetime)

    # This is sent back to the server
    delegatedProxyString = res["Value"]

    ######################################################
    # Equivalent to ProxyManager.completeDelegationUpload

    # Dirty hack:
    X509Chain = get_X509Chain_from_X509Request(x509Req)

    # Create the new chain
    # the pkey was generated together with the Request
    delegatedProxy = X509Chain(keyObj=pkeyReq)
    delegatedProxy.loadChainFromString(delegatedProxyString)

    # make sure the public key match between Request and the new Chain
    # (Stupid, of course it will ! But it is done in the ProxyManager...)
    res = x509Req.checkChain(delegatedProxy)

    assert res["OK"]

    # perform a few checks on the generated proxy

    # There should be one level extra in the delegated proxy
    assert proxyChain.getNumCertsInChain()["Value"] + 1 == delegatedProxy.getNumCertsInChain()["Value"]

    # The issuer of the delegatedProxy should be the original proxy
    assert proxyChain.getCertInChain()["Value"].getSubjectDN() == delegatedProxy.getCertInChain()["Value"].getIssuerDN()

    # The groups should be the same
    assert proxyChain.getDIRACGroup(ignoreDefault=True) == delegatedProxy.getDIRACGroup(ignoreDefault=True)

    assert proxyChain.getNotAfterDate()["Value"] >= delegatedProxy.getNotAfterDate()["Value"]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import os

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities.List import stringListToString, intListToString
from DIRAC.DataManagementSystem.DB.FileCatalogComponents.FileManager.FileManagerBase import FileManagerBase


class FileManagerFlat(FileManagerBase):

    ######################################################
    #
    # The all important _findFiles and _getDirectoryFiles methods
    #

    def _findFiles(self, lfns, metadata=["FileID"], connection=False):
        connection = self._getConnection(connection)
        """ Find file ID if it exists for the given list of LFNs """
        dirDict = self._getFileDirectories(lfns)
        failed = {}
        directoryIDs = {}
        for dirPath in dirDict.keys():
            res = self.db.dtree.findDir(dirPath)
            if not res["OK"] or not res["Value"]:
                error = res.get("Message", "No such file or directory")
                for fileName in dirDict[dirPath]:
                    failed["%s/%s" % (dirPath, fileName)] = error
            else:
                directoryIDs[dirPath] = res["Value"]
        successful = {}
        for dirPath in directoryIDs.keys():
            fileNames = dirDict[dirPath]
            res = self._getDirectoryFiles(directoryIDs[dirPath], fileNames, metadata, connection=connection)
            if not res["OK"] or not res["Value"]:
                error = res.get("Message", "No such file or directory")
                for fileName in fileNames:
                    failed["%s/%s" % (dirPath, fileName)] = error
            else:
                for fileName, fileDict in res["Value"].items():
                    successful["%s/%s" % (dirPath, fileName)] = fileDict
        return S_OK({"Successful": successful, "Failed": failed})

    def _getDirectoryFiles(self, dirID, fileNames, metadata, allStatus=False, connection=False):
        connection = self._getConnection(connection)
        # metadata can be any of
        # ['FileID','Size','UID','GID','Checksum','ChecksumType','Type','CreationDate','ModificationDate','Mode','Status']
        req = "SELECT FileName,%s FROM FC_Files WHERE DirID=%d" % (intListToString(metadata), dirID)
        if not allStatus:
            statusIDs = []
            res = self._getStatusInt("AprioriGood", connection=connection)
            if res["OK"]:
                statusIDs.append(res["Value"])
            if statusIDs:
                req = "%s AND Status IN (%s)" % (req, intListToString(statusIDs))
        if fileNames:
            req = "%s AND FileName IN (%s)" % (req, stringListToString(fileNames))
        res = self.db._query(req, connection)
        if not res["OK"]:
            return res
        files = {}
        for fTuple in res["Value"]:
            fileName = fTuple[0]
            files[fileName] = dict(zip(metadata, fTuple[1:]))
        return S_OK(files)

    ######################################################
    #
    # _addFiles related methods
    #

    def _insertFiles(self, lfns, uid, gid, connection=False):
        connection = self._getConnection(connection)
        # Add the files
        failed = {}
        directoryFiles = {}
        insertTuples = []
        res = self._getStatusInt("AprioriGood", connection=connection)
        statusID = 0
        if res["OK"]:
            statusID = res["Value"]
        for lfn in sorted(lfns.keys()):
            fileInfo = lfns[lfn]
            size = fileInfo["Size"]
            guid = fileInfo.get("GUID", "")
            checksum = fileInfo["Checksum"]
            checksumtype = fileInfo.get("ChecksumType", "Adler32")
            dirName = os.path.dirname(lfn)
            dirID = fileInfo["DirID"]
            fileName = os.path.basename(lfn)
            if dirName not in directoryFiles:
                directoryFiles[dirName] = []
            directoryFiles[dirName].append(fileName)
            insertTuples.append(
                "(%d,%d,%d,%d,%d,'%s','%s','%s','%s',UTC_TIMESTAMP(),UTC_TIMESTAMP(),%d)"
                % (dirID, size, uid, gid, statusID, fileName, guid, checksum, checksumtype, self.db.umask)
            )
        fields = "DirID,Size,UID,GID,Status,FileName,GUID,Checksum,ChecksumType,CreationDate,ModificationDate,Mode"
        req = "INSERT INTO FC_Files (%s) VALUES %s" % (fields, ",".join(insertTuples))
        res = self.db._update(req, connection)
        if not res["OK"]:
            return res
        # Get the fileIDs for the inserted files
        res = self._findFiles(list(lfns), ["FileID"], connection=connection)
        if not res["OK"]:
            for lfn in list(lfns):
                failed[lfn] = "Failed post insert check"
                lfns.pop(lfn)
        else:
            failed.update(res["Value"]["Failed"])
            for lfn, fileDict in res["Value"]["Successful"].items():
                lfns[lfn]["FileID"] = fileDict["FileID"]
        return S_OK({"Successful": lfns, "Failed": failed})

    def _getFileIDFromGUID(self, guid, connection=False):
        connection = self._getConnection(connection)
        if not guid:
            return S_OK({})
        if not isinstance(guid, (list, tuple)):
            guid = [guid]
        req = "SELECT FileID,GUID FROM FC_Files WHERE GUID IN (%s)" % stringListToString(guid)
        res = self.db._query(req, connection)
        if not res["OK"]:
            return res
        guidDict = {}
        for fileID, guid in res["Value"]:
            guidDict[guid] = fileID
        return S_OK(guidDict)

    ######################################################
    #
    # _deleteFiles related methods
    #

    def _deleteFiles(self, fileIDs, connection=False):
        connection = self._getConnection(connection)
        replicaPurge = self.__deleteFileReplicas(fileIDs)
        filePurge = self.__deleteFiles(fileIDs, connection=connection)
        if not replicaPurge["OK"]:
            return replicaPurge
        if not filePurge["OK"]:
            return filePurge
        return S_OK()

    def __deleteFileReplicas(self, fileIDs, connection=False):
        connection = self._getConnection(connection)
        if not fileIDs:
            return S_OK()
        req = "DELETE FROM FC_Replicas WHERE FileID in (%s)" % (intListToString(fileIDs))
        return self.db._update(req, connection)

    def __deleteFiles(self, fileIDs, connection=False):
        connection = self._getConnection(connection)
        if not fileIDs:
            return S_OK()
        req = "DELETE FROM FC_Files WHERE FileID in (%s)" % (intListToString(fileIDs))
        return self.db._update(req, connection)

    ######################################################
    #
    # _addReplicas related methods
    #

    def _insertReplicas(self, lfns, master=False, connection=False):
        connection = self._getConnection(connection)
        res = self._getStatusInt("AprioriGood", connection=connection)
        statusID = 0
        if res["OK"]:
            statusID = res["Value"]
        replicaType = "Replica"
        if master:
            replicaType = "Master"
        insertTuples = {}
        deleteTuples = []
        successful = {}
        failed = {}
        directorySESizeDict = {}
        for lfn in sorted(lfns.keys()):
            fileID = lfns[lfn]["FileID"]
            pfn = lfns[lfn]["PFN"]
            seName = lfns[lfn]["SE"]
            res = self.db.seManager.findSE(seName)
            if not res["OK"]:
                failed[lfn] = res["Message"]
                continue
            seID = res["Value"]
            if not master:
                res = self.__existsReplica(fileID, seID, connection=connection)
                if not res["OK"]:
                    failed[lfn] = res["Message"]
                    continue
                elif res["Value"]:
                    successful[lfn] = True
                    continue
            dirID = lfns[lfn]["DirID"]
            if dirID not in directorySESizeDict:
                directorySESizeDict[dirID] = {}
            if seID not in directorySESizeDict[dirID]:
                directorySESizeDict[dirID][seID] = {"Files": 0, "Size": 0}
            directorySESizeDict[dirID][seID]["Size"] += lfns[lfn]["Size"]
            directorySESizeDict[dirID][seID]["Files"] += 1
            insertTuples[lfn] = "(%d,%d,%d,'%s',UTC_TIMESTAMP(),UTC_TIMESTAMP(),'%s')" % (
                fileID,
                seID,
                statusID,
                replicaType,
                pfn,
            )
            deleteTuples.append((fileID, seID))
        if insertTuples:
            fields = "FileID,SEID,Status,RepType,CreationDate,ModificationDate,PFN"
            req = "INSERT INTO FC_Replicas (%s) VALUES %s" % (fields, ",".join(insertTuples.values()))
            res = self.db._update(req, connection)
            if not res["OK"]:
                self.__deleteReplicas(deleteTuples, connection=connection)
                for lfn in insertTuples.keys():
                    failed[lfn] = res["Message"]
            else:
                # Update the directory usage
                self._updateDirectoryUsage(directorySESizeDict, "+", connection=connection)
                for lfn in insertTuples.keys():
                    successful[lfn] = True
        return S_OK({"Successful": successful, "Failed": failed})

    def __existsReplica(self, fileID, seID, connection=False):
        # TODO: This is in efficient. Should perform bulk operation
        connection = self._getConnection(connection)
        """ Check if a replica already exists """
        if isinstance(seID, six.string_types):
            res = self.db.seManager.findSE(seID)
            if not res["OK"]:
                return res
            seID = res["Value"]
        req = "SELECT FileID FROM FC_Replicas WHERE FileID=%d AND SEID=%d" % (fileID, seID)
        result = self.db._query(req, connection)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK(False)
        return S_OK(True)

    ######################################################
    #
    # _deleteReplicas related methods
    #

    def _deleteReplicas(self, lfns, connection=False):
        connection = self._getConnection(connection)
        successful = {}
        res = self._findFiles(list(lfns), ["DirID", "FileID", "Size"], connection=connection)
        failed = res["Value"]["Failed"]
        lfnFileIDDict = res["Value"]["Successful"]
        toRemove = []
        directorySESizeDict = {}
        for lfn, fileDict in lfnFileIDDict.items():
            fileID = fileDict["FileID"]
            se = lfns[lfn]["SE"]
            toRemove.append((fileID, se))
            # Now prepare the storage usage dict
            res = self.db.seManager.findSE(se)
            if not res["OK"]:
                return res
            seID = res["Value"]
            dirID = fileDict["DirID"]
            if dirID not in directorySESizeDict:
                directorySESizeDict[dirID] = {}
            if seID not in directorySESizeDict[dirID]:
                directorySESizeDict[dirID][seID] = {"Files": 0, "Size": 0}
            directorySESizeDict[dirID][seID]["Size"] += fileDict["Size"]
            directorySESizeDict[dirID][seID]["Files"] += 1
        res = self.__deleteReplicas(toRemove)
        if not res["OK"]:
            for lfn in lfnFileIDDict.keys():
                failed[lfn] = res["Message"]
        else:
            # Update the directory usage
            self._updateDirectoryUsage(directorySESizeDict, "-", connection=connection)
            for lfn in lfnFileIDDict.keys():
                successful[lfn] = True
        return S_OK({"Successful": successful, "Failed": failed})

    def __deleteReplicas(self, replicaTuples, connection=False):
        connection = self._getConnection(connection)
        deleteTuples = []
        for fileID, seID in replicaTuples:
            if isinstance(seID, six.string_types):
                res = self.db.seManager.findSE(seID)
                if not res["OK"]:
                    return res
                seID = res["Value"]
            deleteTuples.append("(%d,%d)" % (fileID, seID))
        req = "DELETE FROM FC_Replicas WHERE (FileID,SEID) IN (%s)" % intListToString(deleteTuples)
        return self.db._update(req, connection)

    ######################################################
    #
    # _setReplicaStatus _setReplicaHost _setReplicaParameter methods
    # _setFileParameter method
    #

    def _setReplicaStatus(self, fileID, se, status, connection=False):
        connection = self._getConnection(connection)
        res = self._getStatusInt(status, connection=connection)
        if not res["OK"]:
            return res
        statusID = res["Value"]
        return self._setReplicaParameter(fileID, se, "Status", statusID, connection=connection)

    def _setReplicaHost(self, fileID, se, newSE, connection=False):
        connection = self._getConnection(connection)
        res = self.db.seManager.findSE(newSE)
        if not res["OK"]:
            return res
        newSE = res["Value"]
        return self._setReplicaParameter(fileID, se, "SEID", newSE, connection=connection)

    def _setReplicaParameter(self, fileID, seID, paramName, paramValue, connection=False):
        connection = self._getConnection(connection)
        if isinstance(seID, six.string_types):
            res = self.db.seManager.findSE(seID)
            if not res["OK"]:
                return res
            seID = res["Value"]
        req = "UPDATE FC_Replicas SET %s='%s', ModificationDate=UTC_TIMESTAMP() WHERE FileID=%d AND SEID=%d;" % (
            paramName,
            paramValue,
            fileID,
            seID,
        )
        return self.db._update(req, connection)

    def _setFileParameter(self, fileID, paramName, paramValue, connection=False):
        connection = self._getConnection(connection)
        if not isinstance(fileID, (list, tuple)):
            fileID = [fileID]
        req = "UPDATE FC_Files SET %s='%s', ModificationDate=UTC_TIMESTAMP() WHERE FileID IN (%s)" % (
            paramName,
            paramValue,
            intListToString(fileID),
        )
        return self.db._update(req, connection)

    ######################################################
    #
    # _getFileReplicas related methods
    #

    def _getFileReplicas(self, fileIDs, fields=["PFN"], connection=False):
        connection = self._getConnection(connection)
        if not fileIDs:
            return S_ERROR("No such file or directory")
        req = "SELECT FileID,SEID,Status,%s FROM FC_Replicas WHERE FileID IN (%s);" % (
            intListToString(fields),
            intListToString(fileIDs),
        )
        res = self.db._query(req, connection)
        if not res["OK"]:
            return res
        replicas = {}
        for fTuple in res["Value"]:
            fileID = fTuple[0]
            if fileID not in replicas:
                replicas[fileID] = {}
            seID = fTuple[1]
            res = self.db.seManager.getSEName(seID)
            if not res["OK"]:
                continue
            seName = res["Value"]
            statusID = fTuple[2]
            res = self._getIntStatus(statusID, connection=connection)
            if not res["OK"]:
                continue
            status = res["Value"]
            replicas[fileID][seName] = {"Status": status}
            replicas[fileID][seName].update(dict(zip(fields, fTuple[3:])))
        for fileID in fileIDs:
            if fileID not in replicas:
                replicas[fileID] = {}
        return S_OK(replicas)



########################################################################
# $HeadURL $
# File: ForwardDISET.py
# Author: Krzysztof.Ciba@NOSPAMgmail.com
# Date: 2013/03/22 12:40:06
########################################################################
""" :mod: ForwardDISET

    ==================

    .. module: ForwardDISET

    :synopsis: DISET forwarding operation handler

    .. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

    DISET forwarding operation handler
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id $"

# #
# @file ForwardDISET.py
# @author Krzysztof.Ciba@NOSPAMgmail.com
# @date 2013/03/22 12:40:22
# @brief Definition of ForwardDISET class.

# # imports
from DIRAC import S_OK, S_ERROR, gConfig
from DIRAC.RequestManagementSystem.private.OperationHandlerBase import OperationHandlerBase
from DIRAC.Core.Base.Client import executeRPCStub
from DIRAC.Core.Utilities import DEncode
from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData

########################################################################


class ForwardDISET(OperationHandlerBase):
    """
    .. class:: ForwardDISET

    functor forwarding DISET operations
    """

    def __init__(self, operation=None, csPath=None):
        """c'tor

        :param Operation operation: an Operation instance
        :param str csPath: CS path for this handler
        """
        # # call base class c'tor
        OperationHandlerBase.__init__(self, operation, csPath)

    # We can ignore the warnings here because we are just
    # replaying something that ought to be sanitized already
    @DEncode.ignoreEncodeWarning
    def __call__(self):
        """execute RPC stub"""
        # # decode arguments
        try:
            decode, length = DEncode.decode(self.operation.Arguments)
            self.log.debug("decoded len=%s val=%s" % (length, decode))
        except ValueError as error:
            self.log.exception(error)
            self.operation.Error = str(error)
            self.operation.Status = "Failed"
            return S_ERROR(str(error))

        # Ensure the forwarded request is done on behalf of the request owner
        decode[0][1]["delegatedDN"] = self.request.OwnerDN
        decode[0][1]["delegatedGroup"] = self.request.OwnerGroup

        # ForwardDiset is supposed to be used with a host certificate
        useServerCertificate = gConfig.useServerCertificate()
        gConfigurationData.setOptionInCFG("/DIRAC/Security/UseServerCertificate", "true")
        forward = executeRPCStub(decode)
        if not useServerCertificate:
            gConfigurationData.setOptionInCFG("/DIRAC/Security/UseServerCertificate", "false")

        if not forward["OK"]:
            self.log.error("unable to execute operation", "'%s' : %s" % (self.operation.Type, forward["Message"]))
            self.operation.Error = forward["Message"]
            return forward
        self.log.info("DISET forwarding done")
        self.operation.Status = "Done"
        return S_OK()



""" Class that collects utilities used in Accounting and Monitoring systems
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC.Core.Utilities import Time


class DBUtils(object):
    def __init__(self, db, setup):
        self._acDB = db
        self._setup = setup

    def _retrieveBucketedData(
        self, typeName, startTime, endTime, selectFields, condDict=None, groupFields=None, orderFields=None
    ):
        """
        Get data from the DB
        Parameters:
          - typeName -> typeName
          - startTime & endTime -> datetime objects. Do I need to explain the meaning?
          - selectFields -> tuple containing a string and a list of fields:
                            ( "SUM(%s), %s/%s", ( "field1name", "field2name", "field3name" ) )
          - condDict -> conditions for the query
                        key -> name of the key field
                        value -> list of possible values
          - groupFields -> list of fields to group by, can be in form
                           ( "%s, %s", ( "field1name", "field2name", "field3name" ) )
          - orderFields -> list of fields to order by, can be in form
                           ( "%s, %s", ( "field1name", "field2name", "field3name" )
        """
        validCondDict = {}
        if isinstance(condDict, dict):
            for key in condDict:
                if isinstance(condDict[key], (list, tuple)) and condDict[key]:
                    validCondDict[key] = condDict[key]
        return self._acDB.retrieveBucketedData(
            self._setup, typeName, startTime, endTime, selectFields, condDict, groupFields, orderFields
        )

    def _getUniqueValues(self, typeName, startTime, endTime, condDict, fieldList):
        stringList = ["%s" for _field in fieldList]
        return self._retrieveBucketedData(
            typeName, startTime, endTime, (",".join(stringList), fieldList), condDict, fieldList
        )

    def _groupByField(self, fieldIndex, dataList):
        """
        From a list of lists/tuples group them into a dict of lists using as key field fieldIndex
        """
        groupDict = {}
        for row in dataList:
            groupingField = row[fieldIndex]
            if groupingField not in groupDict:
                groupDict[groupingField] = []
            if isinstance(row, tuple):
                rowL = list(row[:fieldIndex])
                rowL.extend(row[fieldIndex + 1 :])
                row = rowL
            else:
                del row[fieldIndex]
            groupDict[groupingField].append(row)
        return groupDict

    def _getBins(self, typeName, startTime, endTime):
        return self._acDB.calculateBuckets(self._setup, typeName, startTime, endTime)

    def _getBucketLengthForTime(self, typeName, momentEpoch):
        nowEpoch = Time.toEpoch()
        return self._acDB.calculateBucketLengthForTime(self._setup, typeName, nowEpoch, momentEpoch)

    def _spanToGranularity(self, granularity, bucketsData):
        """
        bucketsData must be a list of lists where each list contains
          - field 0: datetime
          - field 1: bucketLength
          - fields 2-n: numericalFields
        """
        normData = {}

        def addToNormData(bucketDate, data, proportion=1.0):
            if bucketDate in normData:
                for iP in range(len(data)):
                    val = data[iP]
                    if val is None:
                        val = 0
                    normData[bucketDate][iP] += float(val) * proportion
                normData[bucketDate][-1] += proportion
            else:
                normData[bucketDate] = []
                for fD in data:
                    if fD is None:
                        fD = 0
                    normData[bucketDate].append(float(fD) * proportion)
                normData[bucketDate].append(proportion)

        for bucketData in bucketsData:
            bucketDate = bucketData[0]
            originalBucketLength = bucketData[1]
            bucketValues = bucketData[2:]
            if originalBucketLength == granularity:
                addToNormData(bucketDate, bucketValues)
            else:
                startEpoch = bucketDate
                endEpoch = bucketDate + originalBucketLength
                newBucketEpoch = startEpoch - startEpoch % granularity
                if startEpoch == endEpoch:
                    addToNormData(newBucketEpoch, bucketValues)
                else:
                    while newBucketEpoch < endEpoch:
                        start = max(newBucketEpoch, startEpoch)
                        end = min(newBucketEpoch + granularity, endEpoch)
                        proportion = float(end - start) / originalBucketLength
                        addToNormData(newBucketEpoch, bucketValues, proportion)
                        newBucketEpoch += granularity
        return normData

    def _sumToGranularity(self, granularity, bucketsData):
        """
        bucketsData must be a list of lists where each list contains
          - field 0: datetime
          - field 1: bucketLength
          - fields 2-n: numericalFields
        """
        normData = self._spanToGranularity(granularity, bucketsData)
        for bDate in normData:
            del normData[bDate][-1]
        return normData

    def _averageToGranularity(self, granularity, bucketsData):
        """
        bucketsData must be a list of lists where each list contains
          - field 0: datetime
          - field 1: bucketLength
          - fields 2-n: numericalFields
        """
        normData = self._spanToGranularity(granularity, bucketsData)
        for bDate in normData:
            for iP in range(len(normData[bDate])):
                normData[bDate][iP] = float(normData[bDate][iP]) / normData[bDate][-1]
            del normData[bDate][-1]
        return normData

    def _convertNoneToZero(self, bucketsData):
        """
        Convert None to 0
        bucketsData must be a list of lists where each list contains
          - field 0: datetime
          - field 1: bucketLength
          - fields 2-n: numericalFields
        """
        for iPos in range(len(bucketsData)):
            data = bucketsData[iPos]
            for iVal in range(2, len(data)):
                if data[iVal] is None:
                    data[iVal] = 0
        return bucketsData

    def _fillWithZero(self, granularity, startEpoch, endEpoch, dataDict):
        """
        Fill with zeros missing buckets
          - dataDict = { 'key' : { time1 : value,  time2 : value... }, 'key2'.. }
        """
        startBucketEpoch = startEpoch - startEpoch % granularity
        for key in dataDict:
            currentDict = dataDict[key]
            for timeEpoch in range(int(startBucketEpoch), int(endEpoch), granularity):
                if timeEpoch not in currentDict:
                    currentDict[timeEpoch] = 0
        return dataDict

    def _getAccumulationMaxValue(self, dataDict):
        """
        Divide by factor the values and get the maximum value
          - dataDict = { 'key' : { time1 : value,  time2 : value... }, 'key2'.. }
        """
        maxValue = 0
        maxEpoch = 0
        for key in dataDict:
            currentDict = dataDict[key]
            for timeEpoch in currentDict:
                if timeEpoch > maxEpoch:
                    maxEpoch = timeEpoch
                    maxValue = 0
                if timeEpoch == maxEpoch:
                    maxValue += currentDict[timeEpoch]
        return maxValue

    def _getMaxValue(self, dataDict):
        """
        Divide by factor the values and get the maximum value
          - dataDict = { 'key' : { time1 : value,  time2 : value... }, 'key2'.. }
        """
        maxValues = {}
        for key in dataDict:
            currentDict = dataDict[key]
            for timeEpoch in currentDict:
                if timeEpoch not in maxValues:
                    maxValues[timeEpoch] = 0
                maxValues[timeEpoch] += currentDict[timeEpoch]
        maxValue = 0
        for k in maxValues:
            maxValue = max(maxValue, k)
        return maxValue

    def _divideByFactor(self, dataDict, factor):
        """
        Divide by factor the values and get the maximum value
          - dataDict = { 'key' : { time1 : value,  time2 : value... }, 'key2'.. }
        """
        maxValue = 0.0
        for key in dataDict:
            currentDict = dataDict[key]
            for timeEpoch in currentDict:
                currentDict[timeEpoch] /= float(factor)
                maxValue = max(maxValue, currentDict[timeEpoch])
        return dataDict, maxValue

    def _accumulate(self, granularity, startEpoch, endEpoch, dataDict):
        """
        Accumulate all the values.
          - dataDict = { 'key' : { time1 : value,  time2 : value... }, 'key2'.. }
        """
        startBucketEpoch = startEpoch - startEpoch % granularity
        for key in dataDict:
            currentDict = dataDict[key]
            lastValue = 0
            for timeEpoch in range(startBucketEpoch, endEpoch, granularity):
                if timeEpoch in currentDict:
                    lastValue += currentDict[timeEpoch]
                currentDict[timeEpoch] = lastValue
        return dataDict

    def stripDataField(self, dataDict, fieldId):
        """
        Strip <fieldId> data and sum the rest as it was data from one key

        :param dict dataDict: dictionary of the form::

            { 'key' : { <timeEpoch1>: [1, 2, 3],
                        <timeEpoch2>: [3, 4, 5].. } }

          The dataDict is modified in this function and the return structure is:

          .. code-block :: python

                  dataDict : { 'key' : { <timeEpoch1>: 1,
                                         <timeEpoch2>: 3.. } }

        :param int fieldId:

        :returns: list of dictionaries

          .. code-block:: python

                     [ { <timeEpoch1>: 2, <timeEpoch2>: 4... }
                       { <timeEpoch1>: 3, <timeEpoch2>): 5... } ]

        :rtype: python:list

        """
        remainingData = [{}]  # Hack for empty data
        for key in dataDict:
            for timestamp in dataDict[key]:
                for iPos in dataDict[key][timestamp]:
                    remainingData.append({})
                break
            break
        for key in dataDict:
            for timestamp in dataDict[key]:
                strippedField = float(dataDict[key][timestamp][fieldId])
                del dataDict[key][timestamp][fieldId]
                for iPos in range(len(dataDict[key][timestamp])):
                    if timestamp in remainingData[iPos]:
                        remainingData[iPos][timestamp] += float(dataDict[key][timestamp][iPos])
                    else:
                        remainingData[iPos][timestamp] = float(dataDict[key][timestamp][iPos])
                dataDict[key][timestamp] = strippedField

        return remainingData

    def getKeyValues(self, typeName, condDict):
        """
        Get all valid key values in a type
        """
        return self._acDB.getKeyValues(self._setup, typeName, condDict)

    def _calculateProportionalGauges(self, dataDict):
        """
        Get a dict with more than one entry per bucket and list
        """
        bucketSums = {}
        # Calculate total sums in buckets
        for key in dataDict:
            for timeKey in dataDict[key]:
                timeData = dataDict[key][timeKey]
                if len(timeData) < 2:
                    raise Exception(
                        "DataDict must be of the type { <key>:{ <timeKey> : [ field1, field2, ..] } }. With at least two fields"
                    )
                if timeKey not in bucketSums:
                    bucketSums[timeKey] = [0, 0, 0]
                bucketSums[timeKey][0] += timeData[0]
                bucketSums[timeKey][1] += timeData[1]
                bucketSums[timeKey][2] += timeData[0] / timeData[1]
        # Calculate proportionalFactor
        for timeKey in bucketSums:
            timeData = bucketSums[timeKey]
            if bucketSums[timeKey][0] == 0:
                bucketSums[timeKey] = 0
            else:
                bucketSums[timeKey] = (timeData[0] / timeData[1]) / timeData[2]
        # Calculate proportional Gauges
        for key in dataDict:
            for timeKey in dataDict[key]:
                timeData = dataDict[key][timeKey]
                dataDict[key][timeKey] = [(timeData[0] / timeData[1]) * bucketSums[timeKey]]

        return dataDict

    def _getBucketTotals(self, dataDict):
        """
        Sum key data and get totals for each bucket
        """
        newData = {}
        for k in dataDict:
            for bt in dataDict[k]:
                if bt not in newData:
                    newData[bt] = 0.0
                newData[bt] += dataDict[k][bt]
        return newData
